from collections import OrderedDict
from archx.utils import get_prod

def dram_event(width, event_type, architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    
    performance_dict = OrderedDict()

    gigabyte = (2 ** 33) # 1 GB = 2^33 bits
    megabyte = (2 ** 23) # 1 MB = 2^23 bits
    frequency = architecture_dict['dram']['query']['frequency']
    if 'irouter' in architecture_dict:
        bandwidth = architecture_dict['irouter']['query']['bandwidth'] * megabyte
    else:
        bandwidth = architecture_dict['dram']['query']['bandwidth'] * gigabyte

    performance_dict['cycle_count'] = OrderedDict({'value': width / bandwidth, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': (width/bandwidth) / 1000 / frequency, 'unit': 'ms'})
    dram_dict = OrderedDict({'operation': OrderedDict({'dynamic_energy': event_type})})
    performance_dict['subevent'] = OrderedDict({'dram': dram_dict})

    return performance_dict

def dram_input_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    isram_width = architecture_dict['isram']['query']['width']
    return dram_event(isram_width, 'read', architecture_dict, workload_dict)

def dram_weight_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    wsram_width = architecture_dict['wsram']['query']['width']
    return dram_event(wsram_width, 'read', architecture_dict, workload_dict=None)

def dram_output_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    osram_width = architecture_dict['osram']['query']['width']
    return dram_event(osram_width, 'read', architecture_dict, workload_dict=None)

def dram_output_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:
    osram_width = architecture_dict['osram']['query']['width']
    return dram_event(osram_width, 'write', architecture_dict, workload_dict=None)

def sram_event(event_type, sram, architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    frequency = architecture_dict[sram]['query']['frequency']

    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    events = 1 / router_dim

    performance_dict['cycle_count'] = OrderedDict({'value': events, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': events / 1000 / frequency, 'unit': 'ms'})
    sram_dict = OrderedDict({'operation': OrderedDict({'dynamic_energy': event_type})})
    performance_dict['subevent'] = OrderedDict({sram: sram_dict})

    return performance_dict

def isram_offchip_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('read', 'isram', architecture_dict, workload_dict)

def isram_offchip_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('write', 'isram', architecture_dict, workload_dict)

def wsram_offchip_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('read', 'wsram', architecture_dict, workload_dict)

def wsram_offchip_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('write', 'wsram', architecture_dict, workload_dict)

def osram_offchip_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('read', 'osram', architecture_dict, workload_dict)

def osram_offchip_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('write', 'osram', architecture_dict, workload_dict)

def isram_onchip_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('read', 'isram', architecture_dict, workload_dict)

def isram_onchip_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('write', 'isram', architecture_dict, workload_dict)

def wsram_onchip_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('read', 'wsram', architecture_dict, workload_dict)

def wsram_onchip_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('write', 'wsram', architecture_dict, workload_dict)

def osram_onchip_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('read', 'osram', architecture_dict, workload_dict)

def osram_onchip_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    return sram_event('write', 'osram', architecture_dict, workload_dict)


from loguru import logger

class Event:
    def __init__(self):
        self.events = {'event': {}}

    def add_event(self, name: str, subevent: list, performance: str):
        assert isinstance(name, str), "'name' parameter must be of type 'str'."
        assert name not in self.events, f"Event '{name}' already exists in event."
        assert isinstance(subevent, list), "'subevent' parameter must be of type 'list'."
        assert isinstance(performance, str), "'performance' parameter must be of type 'str' or not provided."

        self.events['event'][name] = {'subevent': subevent, 'performance': performance}

        logger.info(f"Added event: {name}")
        for sub in subevent:
            logger.debug(f"\tSubevent: {sub}")
        logger.debug(f"\tPerformance: {performance}")

    def to_yaml(self):
        return self.events


from archx.metric import query_module_metric, aggregate_event_metric, aggregate_tag_metric, aggregate_event_count
from collections import OrderedDict
from archx.utils import get_prod, read_yaml
from archx.architecture import load_architecture_dict
from archx.event import load_event_graph
from archx.metric import load_metric_dict
import statistics
import numpy as np

def geomean(dict_list: list[OrderedDict]) -> OrderedDict:
    geomean_dict = OrderedDict()

    for dict in dict_list:
        for key, value in dict.items():
            if key not in geomean_dict:
                if isinstance(value, str):
                    geomean_dict[key] = value
                else:
                    geomean_dict[key] = []

            if not isinstance(value, str):
                geomean_dict[key].append(value)

    for key, value in geomean_dict.items():
        if isinstance(value, list):
            geomean_dict[key] = statistics.geometric_mean(value)

    return geomean_dict

def query_cycle_count(event_graph, metric_dict, workload, event) -> OrderedDict:
    cycle_count_dict = aggregate_event_metric(event_graph=event_graph, metric_dict=metric_dict, metric='cycle_count', workload=workload, event=event)
    return cycle_count_dict['value']

def query_execution_time(event_graph, metric_dict, workload, event) -> OrderedDict:
    execution_time_dict = aggregate_event_metric(event_graph=event_graph, metric_dict=metric_dict, metric='runtime', workload=workload, event=event)
    return execution_time_dict['value'] / 10**3 # ms -> s

def query_dynamic_energy(event_graph, metric_dict, workload, event=None, tag=None) -> OrderedDict:
    if tag is None:
        dynamic_energy_dict = aggregate_event_metric(event_graph=event_graph, metric_dict=metric_dict, metric='dynamic_energy', workload=workload, event=event)
    else:
        dynamic_energy_dict = aggregate_tag_metric(event_graph=event_graph, metric_dict=metric_dict, metric='dynamic_energy', workload=workload, tag=tag)
    return dynamic_energy_dict['value'] / 10**9 # nJ -> J

def query_leakage_power(event_graph, metric_dict, workload, event=None, tag=None) -> OrderedDict:
    if tag is None:
        leakage_power = aggregate_event_metric(event_graph=event_graph, metric_dict=metric_dict, metric='leakage_power', workload=workload, event=event)
    else:
        leakage_power =  aggregate_tag_metric(event_graph=event_graph, metric_dict=metric_dict, metric='leakage_power', workload=workload, tag=tag)
    return leakage_power['value'] / 10**3 # mW -> W

def query_area(event_graph, metric_dict, workload=None, tag=None, module=None) -> np.float64:

    if module is not None:
        area = aggregate_event_metric(event_graph=event_graph, metric_dict=metric_dict, metric='area', workload=workload, event=module)['value']

    elif tag is not None:
        area = aggregate_tag_metric(event_graph=event_graph, metric_dict=metric_dict, metric='area', workload=workload, tag=tag)['value']

    return area

def query_operational_carbon(tag, event_graph, metric_dict, workload, event, CI) -> OrderedDict:
    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=workload, event=event)
    dynamic_energy = query_dynamic_energy(event_graph=event_graph, metric_dict=metric_dict, workload=workload, tag=tag)
    leakage_power = query_leakage_power(event_graph=event_graph, metric_dict=metric_dict, workload=workload, tag=tag)
    power = (leakage_power + (dynamic_energy / execution_time)) * 10**3 # W -> mW

    op_carbon = CI * (power / 1000000) * (execution_time / 3600) # mW -> KW, s -> H

    return op_carbon

def query_tag_power(tag, event_graph, metric_dict, workload, event) -> OrderedDict:
    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=workload, event=event)
    dynamic_energy = query_dynamic_energy(event_graph=event_graph, metric_dict=metric_dict, workload=workload, tag=tag)
    leakage_power = query_leakage_power(event_graph=event_graph, metric_dict=metric_dict, workload=workload, tag=tag)
    power = (leakage_power + (dynamic_energy / execution_time)) * 10**3 # W -> mW

    return power

def compute_throughput(performance_metrics_dict: OrderedDict) -> OrderedDict:

    flops = performance_metrics_dict['flops']
    execution_time = performance_metrics_dict['execution_time']
    throughput = flops/execution_time
    performance_metrics_dict['throughput'] = throughput

    return performance_metrics_dict

def compute_latency(performance_metrics_dict: OrderedDict) -> OrderedDict:

    flops = performance_metrics_dict['flops']

    execution_time = performance_metrics_dict['execution_time']

    latency = execution_time/flops

    performance_metrics_dict['latency'] = latency


    return performance_metrics_dict

def compute_throughput_efficiancy(performance_metrics_dict: OrderedDict) -> OrderedDict:

    flops = performance_metrics_dict['flops']
    energy = performance_metrics_dict['energy']
    power = performance_metrics_dict['power']
    execution_time = performance_metrics_dict['execution_time']

    throughput = flops/execution_time
    energy_efficiancy = throughput / (energy + (power * execution_time))
    power_efficiancy = throughput / (power + (energy / execution_time))
    
    performance_metrics_dict['throughput'] = throughput
    performance_metrics_dict['energy_efficiency'] = energy_efficiancy
    performance_metrics_dict['power_efficiency'] = power_efficiancy

    return performance_metrics_dict

def query_throughput_metrics(event_graph, metric_dict, module, workload, event)-> OrderedDict:
    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)
    pe_count = aggregate_event_count(event_graph=event_graph, workload=event, event=module)

    flops = pe_count * 2 / 10**9 # GFLOPS

    performance_metrics_dict = OrderedDict({
        'flops': flops,
        'execution_time': execution_time
    })

    return performance_metrics_dict

def query_performance_metrics(event_graph, metric_dict, module, workload, event) -> OrderedDict:

    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)
    cycle_count = query_cycle_count(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)
    pe_count = aggregate_event_count(event_graph=event_graph, workload=event, event=module)
    dynamic_energy = query_dynamic_energy(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')
    leakage_power = query_leakage_power(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')

    flops = pe_count * 2 / 10**9 # GFLOPS

    performance_metrics_dict = OrderedDict({
        'flops': flops,
        'execution_time': execution_time,
        'energy': dynamic_energy,
        'power': leakage_power,
        'cycle_count': cycle_count,
    })

    return performance_metrics_dict

def query_throughput_energy_metrics_workload(event_graph, metric_dict, module, workload, event) -> OrderedDict:

    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)
    pe_count = aggregate_event_count(event_graph=event_graph, workload=event, event=module)
    dynamic_energy = query_dynamic_energy(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')

    flops = pe_count * 2 / 10**9 # GFLOPS

    performance_metrics_dict = OrderedDict({
        'flops': flops,
        'execution_time': execution_time,
        'energy': dynamic_energy,
    })

    return performance_metrics_dict

def query_throughput_energy_metrics(event_graph, metric_dict, module, workload, event) -> OrderedDict:

    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)
    pe_count = aggregate_event_count(event_graph=event_graph, workload=event, event=module)
    dynamic_energy = query_dynamic_energy(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')

    flops = pe_count * 2 / 10**9 # GFLOPS

    performance_metrics_dict = OrderedDict({
        'flops': flops,
        'execution_time': execution_time,
        'energy': dynamic_energy,
    })

    return performance_metrics_dict

def query_performance_gemm_metrics(event_graph, metric_dict, module, workload, event) -> OrderedDict:

    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)
    pe_count = aggregate_event_count(event_graph=event_graph, workload=event, event=module)
    dynamic_energy = query_dynamic_energy(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')
    leakage_power = query_leakage_power(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')

    flops = pe_count * 2 / 10**9 # GFLOPS

    performance_metrics_dict = OrderedDict({
        'flops': flops,
        'execution_time': execution_time,
        'energy': dynamic_energy,
        'power': leakage_power,
    })

    return performance_metrics_dict

def query_performance_nonlinear_metrics(event_graph, metric_dict, module, workload, event) -> OrderedDict:

    execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)
    pe_count = aggregate_event_count(event_graph=event_graph, workload=event, event=module)
    dynamic_energy = query_dynamic_energy(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')
    leakage_power = query_leakage_power(event_graph=event_graph, metric_dict=metric_dict, workload=event, tag='onchip')
    cycle_count = query_cycle_count(event_graph=event_graph, metric_dict=metric_dict, workload=event, event=event)

    flops = pe_count * 3 / 10**9 # GFLOPS

    performance_metrics_dict = OrderedDict({
        'flops': flops,
        'execution_time': execution_time,
        'energy': dynamic_energy,
        'power': leakage_power,
        'cycle_count': cycle_count,
    })

    return performance_metrics_dict

def load_yaml(path):
    yaml_dict = OrderedDict({
        'architecture_dict': load_architecture_dict(path + '/architecture.yaml'),
        'event_graph': load_event_graph(path + '/checkpoint.gt'),
        'metric_dict': load_metric_dict(path + '/metric.yaml'),
        'event_dict': read_yaml(path + '/event.yaml'),
    })

    return yaml_dict


from archx.utils import read_yaml
import itertools, os, copy

def main():
    base_path = 'zoo/llm/'
    scripts_path = base_path + 'scripts/'
    architecture_list = read_yaml(scripts_path + 'architecture_configuration.yaml')

    with open(scripts_path + 'runs.txt', 'w') as file:
        for arch in architecture_list['architecture']:
            arch_path = arch + '/'
            architecture_config_dict = read_yaml(base_path + arch_path + 'architecture/template/configuration.yaml')
            network_dict = architecture_config_dict['network']
            array_dict = architecture_config_dict['array']
            subarch_dict = architecture_config_dict.get('subarchitecture', {'': {}})
            for network_key, network_value in network_dict.items():
                network_path = network_key + '/'
                for network_dim in network_value['dimension']:
                    network_dim_path = network_key + '_' + str(network_dim[0]) + 'x' + str(network_dim[1]) + '/' if len(network_dim) == 2 else network_key + '/'
                    for array_key in array_dict.keys():
                        array_path = array_key + '/'
                        for subarch_key in subarch_dict.keys():
                            subarch_path = subarch_key + '/'
                            workload_config_dict = read_yaml(base_path + arch_path + 'workload/template/configuration.yaml')
                            workload_dict = read_yaml(base_path + arch_path + 'workload/template/workload.yaml')
                            
                            for model in workload_dict['workload'].keys():
                                model_path = model + '/'
                                filtered_workload_config_dict = copy.deepcopy(workload_config_dict)
                                model_key_set = set(workload_dict['workload'][model]['configuration'].keys())
                                config_combination_key_set = set(workload_config_dict['configuration'].keys())
                                shared_keys = set(model_key_set) & set(config_combination_key_set)
                                for shared_key in shared_keys:
                                    del filtered_workload_config_dict['configuration'][shared_key]
                                config_combinations = [dict(zip(filtered_workload_config_dict['configuration'].keys(), combination)) for combination in itertools.product(*filtered_workload_config_dict['configuration'].values())]
                                for config_combination in config_combinations:
                                    workload_config_path = ''
                                    for key, value in config_combination.items():
                                        if len(filtered_workload_config_dict['configuration'][key]) > 1:
                                            workload_config_path += '/' + key + '_' + str(value)
                                    workload_config_path += '/'

                                    for termination in ['full_termination', 'early_termination']:
                                        if arch != 'mugi' and termination == 'early_termination':
                                            continue
                                        elif arch == 'mugi' and termination == 'full_termination':
                                            workload_termination_path = 'full_termination/'
                                            runs_termination_path = 'full_termination/'
                                        elif arch == 'mugi' and termination == 'early_termination':
                                            workload_termination_path = 'early_termination_' + array_path
                                            runs_termination_path = 'early_termination/'
                                        else:
                                            workload_termination_path = ''
                                            runs_termination_path = ''
                                    
                                        if network_key == 'multi_node' and subarch_key in ['pwl', 'taylor']:
                                            continue
                                        if network_key == 'multi_node' and config_combination['batch_size'] !=8:
                                            continue
                                        if network_key == 'multi_node' and config_combination['max_seq_len'] != 4096:
                                            continue
                                        if network_key == 'multi_node' and array_key in ['32x32', '64x64']:
                                            continue
                                        if array_key in ['32x8', '4x4']:
                                            continue
                                        if array_key != '16x16' and subarch_key in ['pwl', 'taylor']:
                                            continue

                                        event_subarch = arch + '/' if subarch_path == '/' else subarch_path

                                        architecture_path = os.path.normpath('./' + base_path + arch_path + '/architecture/generated/' + subarch_path + network_dim_path + array_path + 'architecture.yaml')
                                        event_path = os.path.normpath('./' + base_path + arch_path + 'event/generated/' + network_path + event_subarch + 'event.yaml')
                                        metric_path = os.path.normpath('./' + base_path + 'common/metric/metric.yaml')
                                        runs_path = os.path.normpath('./' + base_path + 'runs/' + arch_path + network_dim_path + subarch_path + array_path + model_path + workload_config_path + runs_termination_path)
                                        workload_path = os.path.normpath('./' + base_path + arch_path + 'workload/generated/' + model_path + workload_termination_path + workload_config_path + 'workload.yaml')
                                        checkpoint_path = os.path.normpath(runs_path + '/checkpoint.gt/')

                                        file.write('-a ' + architecture_path + ' -e ' + event_path + ' -m ' + metric_path + ' -r ' + runs_path +  ' -w ' + workload_path  + ' -c ' + checkpoint_path + ' -s ' + '\n')
    
if '__main__' == __name__:
    main()


import csv, os, copy

from collections import OrderedDict
from loguru import logger

from archx.utils import get_path, interpolate_oneD_linear, interpolate_oneD_quadratic


skip_list = ['technology', 'frequency', 'interpolation', 'dynamic_uw', 'leakage_uw', 'area_jj']
interpolation_list = ['linear', 'quadratic']


def query(name: str, interface: str, query: OrderedDict, input_dir=None, output_dir=None):
    query = copy.deepcopy(query)
    query_class = query['class'].lower()
    query_technology = query['technology']
    query_frequency = query['frequency']
    del query['class']
    del query['technology']
    del query['frequency']

    csv_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'include/csv')
    csv_file = get_path(os.path.join(csv_dir, query_class + '.csv'))

    with open(csv_file) as csv_file:
        reader = csv.DictReader(csv_file)
        technology_flag = False
        for row in reader:
            interpolation = row['interpolation'].lower()

            # no technology scaling
            if row['technology'] == str(query_technology):
                technology_flag = True
                dynamic_power_mW = float(row['dynamic_uw']) # in uw
                leakage_power_mW = float(row['leakage_uw']) # in uw
                dynamic_power_mW /= 1000  # convert to mW
                leakage_power_mW /= 1000  # convert to mW
                area_jj = float(row['area_jj']) # in josephson junctions

                # frequency scaling (dynamic power)
                dynamic_power_mW *= (query_frequency / float(row['frequency'])) # in mW

                for attr_key in query:
                    assert attr_key in row, logger.error(f'Invalid query attribute <{attr_key}>.')

                for attr_key in row:
                    if attr_key not in skip_list:
                        if attr_key not in query:
                            query[attr_key] = 1
                        if interpolation == 'linear':
                            dynamic_power_mW   = interpolate_oneD_linear(    query[attr_key], [{'x': 0, 'y': 0}, {'x': float(row[attr_key]), 'y': dynamic_power_mW}])
                            leakage_power_mW   = interpolate_oneD_linear(    query[attr_key], [{'x': 0, 'y': 0}, {'x': float(row[attr_key]), 'y': leakage_power_mW}])
                            area_jj            = interpolate_oneD_linear(    query[attr_key], [{'x': 0, 'y': 0}, {'x': float(row[attr_key]), 'y': area_jj}])
                        elif interpolation == 'quadratic':
                            dynamic_power_mW   = interpolate_oneD_quadratic( query[attr_key], [{'x': 0, 'y': 0}, {'x': float(row[attr_key]), 'y': dynamic_power_mW}])
                            leakage_power_mW   = interpolate_oneD_quadratic( query[attr_key], [{'x': 0, 'y': 0}, {'x': float(row[attr_key]), 'y': leakage_power_mW}])
                            area_jj            = interpolate_oneD_quadratic( query[attr_key], [{'x': 0, 'y': 0}, {'x': float(row[attr_key]), 'y': area_jj}])
                        else:
                            assert interpolation in interpolation_list, logger.error(f'Invalid interpolation <{interpolation}> in <{csv_file}>. Valid options are <{str(interpolation_list)}>.')
        assert technology_flag is True, logger.error(f'Invalid technology: <{query_technology}>.')

    energy_per_event_nJ = dynamic_power_mW / query_frequency

    output_dict = OrderedDict()
    output_dict['technology'] = OrderedDict({'value': query_technology, 'unit': 'nm'})
    output_dict['frequency'] = OrderedDict({'value': query_frequency, 'unit': 'GHz'})
    output_dict['dynamic_energy'] = OrderedDict({'value': energy_per_event_nJ, 'unit': 'nJ'})
    output_dict['leakage_power'] = OrderedDict({'value': leakage_power_mW, 'unit': 'mW'})
    output_dict['area'] = OrderedDict({'value': area_jj, 'unit': 'jj'})

    return output_dict




import os, sys, yaml, json
import numpy as np

from collections import OrderedDict
from yamlordereddictloader import SafeDumper
from yamlordereddictloader import SafeLoader
from loguru import logger


class bcolors:
    """
    default color palette
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    red = "#EF553B"
    orange = "#E58606"
    yellow = "#FABD2F"
    green = "#9CC424"
    cyan = '#6FD19F'
    blue = "#FABD2F"
    purple = "#AB82FF"
    gray = "#CCCCCC"
    gray2 = "#999999"
    gray3 = "#666666"
    gray4 = "#333333"

    
    ResetAll = "\033[0m"
    Bold       = "\033[1m"
    Dim        = "\033[2m"
    Underlined = "\033[4m"
    Blink      = "\033[5m"
    Reverse    = "\033[7m"
    Hidden     = "\033[8m"

    ResetBold       = "\033[21m"
    ResetDim        = "\033[22m"
    ResetUnderlined = "\033[24m"
    ResetBlink      = "\033[25m"
    ResetReverse    = "\033[27m"
    ResetHidden     = "\033[28m"

    Default      = "\033[39m"
    Black        = "\033[30m"
    Red          = "\033[31m"
    Green        = "\033[32m"
    Yellow       = "\033[33m"
    Blue         = "\033[34m"
    Magenta      = "\033[35m"
    Cyan         = "\033[36m"
    LightGray    = "\033[37m"
    DarkGray     = "\033[90m"
    LightRed     = "\033[91m"
    LightGreen   = "\033[92m"
    LightYellow  = "\033[93m"
    LightBlue    = "\033[94m"
    LightMagenta = "\033[95m"
    LightCyan    = "\033[96m"
    White        = "\033[97m"

    BackgroundDefault      = "\033[49m"
    BackgroundBlack        = "\033[40m"
    BackgroundRed          = "\033[41m"
    BackgroundGreen        = "\033[42m"
    BackgroundYellow       = "\033[43m"
    BackgroundBlue         = "\033[44m"
    BackgroundMagenta      = "\033[45m"
    BackgroundCyan         = "\033[46m"
    BackgroundLightGray    = "\033[47m"
    BackgroundDarkGray     = "\033[100m"
    BackgroundLightRed     = "\033[101m"
    BackgroundLightGreen   = "\033[102m"
    BackgroundLightYellow  = "\033[103m"
    BackgroundLightBlue    = "\033[104m"
    BackgroundLightMagenta = "\033[105m"
    BackgroundLightCyan    = "\033[106m"
    BackgroundWhite        = "\033[107m"


def strip_list(input_list: list) -> list:
    """
    strip leading and trailing spaces for each list item
    """
    l = []

    for e in input_list:
        e = e.strip()
        if e != '' and e != ' ':
            l.append(e)

    return l


def check_type(input, type) -> None:
    """
    check whether input is the required type
    """
    assert isinstance(input, type), logger.error(f'Invalid type <{type}>.')
    

def check_file_list(file_list: list) -> None:
    """
    check whether all files in the list exist
    """
    for file in file_list:
        assert os.path.exists(file), logger.error(f'Invalid file <{file}>.')


def clean_file_list(file_list: list) -> None:
    """
    delete files in the list, if they exist
    """
    for file in file_list:
        if os.path.exists(file):
            logger.warning(f'Delete file <{file}>.')
            os.remove(file)


def create_dir(directory: str) -> None:
    """
    Checks the existence of a directory, if does not exist, create a new one
    :param directory: path to directory under concern
    :return: None
    """
    try:
        if not os.path.exists(directory):
            os.makedirs(directory)
            logger.success(f'Create directory <{directory}>.')
    except OSError:
        logger.error(f'Create directory <{directory}>.')
        sys.exit()
    

def create_subdir(path: str, subdir_list: list) -> None:
    for subdir in subdir_list:
        subdir_path = os.path.join(path, subdir.strip('/'))
        if not os.path.exists(subdir_path):
            create_dir(subdir_path)

    
def read_yaml(file: str) -> OrderedDict:
    return yaml.load(open(file), Loader=SafeLoader)


def write_yaml(file: str, content: str) -> None:
    """
    if file exists at filepath, overwite the file, if not, create a new file
    :param filepath: string that specifies the destination file path
    :param content: yaml string that needs to be written to the destination file
    :return: None
    """
    if os.path.exists(file):
        os.remove(file)
    create_dir(os.path.dirname(file))
    out_file = open(file, 'a')
    out_file.write(yaml.dump( content, default_flow_style= False, Dumper=SafeDumper))


def check_repeated_key(full_dict: OrderedDict, key:str, val: OrderedDict) -> tuple:
    key_index = list(full_dict.keys()).index(key)
    key_list = list(full_dict.keys())[0 : key_index]
    for key in key_list:
        if full_dict[key] == val:
            return True, key
    return False, None
    

# the following interpolate_oneD_linear and interpolate_oneD_quadratic are adapted from accelergy
# ===============================================================
# useful helper functions that are commonly used in estimators
# ===============================================================
def interpolate_oneD_linear(desired_x, known):
    """
    utility function that performs 1D linear interpolation with a known energy value
    :param desired_x: integer value of the desired attribute/argument
    :param known: list of dictionary [{x: <value>, y: <energy>}]
    :return energy value with desired attribute/argument
    """
    # assume E = ax + c where x is a hardware attribute
    ordered_list = []
    if known[1]['x'] < known[0]['x']:
        ordered_list.append(known[1])
        ordered_list.append(known[0])
    else:
        ordered_list = known

    slope = (known[1]['y'] - known[0]['y']) / (known[1]['x'] - known[0]['x'])
    desired_energy = slope * (desired_x - ordered_list[0]['x']) + ordered_list[0]['y']
    return desired_energy


def interpolate_oneD_quadratic(desired_x, known):
    """
    utility function that performs 1D linear interpolation with a known energy value
    :param desired_x: integer value of the desired attribute/argument
    :param known: list of dictionary [{x: <value>, y: <energy>}]
    :return energy value with desired attribute/argument
    """
    # assume E = ax^2 + c where x is a hardware attribute
    ordered_list = []
    if known[1]['x'] < known[0]['x']:
        ordered_list.append(known[1])
        ordered_list.append(known[0])
    else:
        ordered_list = known

    slope = (known[1]['y'] - known[0]['y']) / (known[1]['x']**2 - known[0]['x']**2)
    desired_energy = slope * (desired_x**2 - ordered_list[0]['x']**2) + ordered_list[0]['y']
    return desired_energy


def get_input_tuple(input: tuple | int, size: int=2) -> tuple:
    if isinstance(input, tuple):
        assert len(input) == size, logger.error(f'Invalid size <{str(len(input))}> != <{str(size)}>.')
        return input
    else:
        output = (input, ) * size
        return output


def get_path(path: str, check_exist: bool=True) -> str:
    path = os.path.abspath(path)
    path = os.path.realpath(path)
    if check_exist:
        assert os.path.exists(path), logger.error(f'Invalid path <{path}>')
    return path


def uniquify_list(sequence: list) -> list:
    seen = set()
    return [x for x in sequence if not (x in seen or seen.add(x))]


def get_dict(input_dict: OrderedDict) -> OrderedDict:
    return json.loads(json.dumps(input_dict))


def check_dict_in_list(input_dict: OrderedDict, input_list: list) -> OrderedDict:
    return get_dict(input_dict) in input_list


def check_dict_equal(input_dict0: OrderedDict, input_dict1: OrderedDict) -> bool:
    return get_dict(input_dict0) == get_dict(input_dict1)


def get_prod(input_array: np.ndarray) -> np.ndarray:
    return np.prod(np.array(input_array))




# following two lines are used in testing
import sys, os, shutil

from loguru import logger

from archx.interface import query_interface, register_interface, unregister_interface, copy_interface
from archx.utils import get_path, create_dir


def test_query_interface_csv_cmos():
    module = 'ireg'
    query = {
        'class': 'multiplier_reg',
        'interface': 'csv_cmos',
        'technology': 32,
        'frequency': 400,
        'bitwidth': 32
    }
    output = query_interface(module, query)
    logger.info('test_query_interface_csv_cmos: ', output)


def test_query_interface_cacti7_sram():
    module = 'isram'
    query = {
        'class': 'sram',
        'interface': 'cacti7',
        'technology': 32,
        'frequency': 400,
        'width': 16,
        'depth': 512,
        'bank': 32
    }
    path = get_path('tests')
    path = path + '/test_interface/'
    create_dir(path)
    output = query_interface(module, query, output_dir=path)
    logger.info('test_query_interface_cacti7_sram: ', output)


def test_query_interface_cacti7_dram():
    module = 'ocdram'
    query = {
        'class': 'dram',
        'interface': 'cacti7',
        'technology': 32,
        'frequency': 400,
        'bandwidth': 25.6,
        'size': 1073741824 # 1GB in bytes
    }
    path = get_path('tests')
    path = path + '/test_interface/'
    create_dir(path)
    output = query_interface(module, query, output_dir=path)
    logger.info('test_query_interface_cacti7_dram: ', output)


def test_copy_interface():
    name = 'csv_cmos'
    path = 'tests/test_interface/dummy_csv_cmos'
    copy_interface(name, path)


def test_register_interface():
    name = 'dummy_csv_cmos'
    path = 'tests/test_interface/dummy_csv_cmos'
    register_interface(name, path)


def test_unregister_interface():
    name = 'dummy_csv_cmos'
    unregister_interface(name)


def test_cleanup():
    path = 'tests/test_interface/dummy_csv_cmos'
    shutil.rmtree(path)
    path = get_path('tests/test_interface/')
    shutil.rmtree(path)


if __name__ == "__main__":
    test_query_interface_csv_cmos()
    test_query_interface_cacti7_sram()
    test_query_interface_cacti7_dram()
    test_copy_interface()
    test_register_interface()
    test_unregister_interface()
    test_cleanup()




# following two lines are used in testing
import sys, os, shutil

from loguru import logger

from archx.workload import create_workload_dict, save_workload_dict, load_workload_dict
from archx.utils import get_path, check_dict_equal


def test_create_workload_dict():
    workload_file = 'examples/mac_1_cycle/input/workload/example.workload.yaml'
    workload_dict = create_workload_dict(workload_file)
    for module in workload_dict:
        logger.info(module, workload_dict[module], '\n')

    save_workload_dict(workload_dict, 'tests/test_workload/example.workload_dict.yaml')
    workload_dict_loaded = load_workload_dict('tests/test_workload/example.workload_dict.yaml')
    assert check_dict_equal(workload_dict, workload_dict_loaded)


def test_cleanup():
    path = get_path('tests/test_workload/')
    shutil.rmtree(path)


if __name__ == "__main__":
    test_create_workload_dict()
    test_cleanup()




from .metric import create_metric_dict, save_metric_dict, create_event_metrics, create_module_metrics, aggregate_event_count, aggregate_event_metric, query_module_metric, aggregate_tag_metric, load_metric_dict



from collections import OrderedDict
from loguru import logger
import math

def sum_subevents(performance_dict_1: OrderedDict, performance_dict_2) -> OrderedDict:

    assert performance_dict_1.keys() == performance_dict_2.keys(), logger.error(f'performance_dicts must have the same keys to sum')
    assert performance_dict_1['subevent'].keys() == performance_dict_2['subevent'].keys(), logger.error(f'performance dict subevents must have the same keys to sum')

    sum_performance_dict = OrderedDict()

    for key, value in performance_dict_1.items():
        sum_performance_dict[key] = OrderedDict()
        for subkey, subvalue in value.items():
            if isinstance(subvalue, str):
                sum_performance_dict[key][subkey] = subvalue

    for subevent, subevent_dict in performance_dict_1['subevent'].items():
        sum_performance_dict['subevent'][subevent] = OrderedDict()

        
        for metric, value in subevent_dict.items():
            if not isinstance(value, dict):
                sum_performance_dict['subevent'][subevent][metric] = performance_dict_1['subevent'][subevent][metric] + performance_dict_2['subevent'][subevent][metric]
            else:
                performance_dict_1_count_average = 0 if sum_performance_dict['subevent'][subevent]['count'] == 0 else performance_dict_1['subevent'][subevent]['count'] / sum_performance_dict['subevent'][subevent]['count']
                performance_dict_2_count_average = 0 if sum_performance_dict['subevent'][subevent]['count'] == 0 else performance_dict_2['subevent'][subevent]['count'] / sum_performance_dict['subevent'][subevent]['count']
                sum_performance_dict['subevent'][subevent][metric] = OrderedDict()
                for submetric, subvalue in value.items():
                    sum_performance_dict['subevent'][subevent][metric][submetric] = (performance_dict_1['subevent'][subevent][metric][submetric] * performance_dict_1_count_average) + (performance_dict_2['subevent'][subevent][metric][submetric] * performance_dict_2_count_average)

    return sum_performance_dict

class TiledGEMM:
    """
    Class that tiles a GEMM of two input matrices, given the size of the matrices and size of each tile.
    Handles partial tiling, and computes memory sizes for matrices and tiles.
    This class simulates dimensions of matrices, tiles, and combination of dimensions, not a populated instantiated tile.
    """
    def __init__(self, batch, m, k, n, tile_m, tile_k, tile_n, m_k_bitwidth, k_n_bitwidth, m_n_bitwidth, array_width=None, array_height=None, array_depth=None):

        if 0 in (m, k, n, tile_m, tile_k, tile_n):
            self.is_valid = False
            return

        self.is_valid = True

        #initialize
        self.batch = batch
        self.m = m
        self.k = k
        self.n = n
        self.tile_m = min(tile_m, self.m)
        self.tile_k = min(tile_k, self.k)
        self.tile_n = min(tile_n, self.n)
        self.m_k_bitwidth = m_k_bitwidth
        self.k_n_bitwidth = k_n_bitwidth
        self.m_n_bitwidth = m_n_bitwidth
        self.k_util = self.tile_k / array_depth if array_depth else 1
        self.m_util = self.tile_m / array_width if array_width else 1
        self.n_util = self.tile_n / array_height if array_height else 1
        
        # Batch Dims
        self.m_total = self.m * self.batch
        self.k_total = self.k * self.batch
        self.n_total = self.n * self.batch

        #Total Tiles
        self.m_tiles = math.ceil(self.m / self.tile_m)
        self.k_tiles = math.ceil(self.k / self.tile_k)
        self.n_tiles = math.ceil(self.n / self.tile_n)
        
        # Batch Tiles
        self.m_total_tiles = self.m_tiles * self.batch
        self.k_total_tiles = self.k_tiles * self.batch
        self.n_total_tiles = self.n_tiles * self.batch        

        # GEMM Tiles
        self.m_k_matrix_tiles = self.m_tiles * self.k_tiles
        self.k_n_matrix_tiles = self.k_tiles * self.n_tiles
        self.m_n_matrix_tiles = self.m_tiles * self.n_tiles
        self.m_k_n_matrix_tiles = self.m_tiles * self.k_tiles * self.n_tiles

        # GEMM tiles across batch
        self.m_k_total_tiles = self.m_k_matrix_tiles * self.batch
        self.k_n_total_tiles = self.k_n_matrix_tiles * self.batch
        self.m_n_total_tiles = self.m_n_matrix_tiles * self.batch
        self.m_k_n_total_tiles = self.m_k_n_matrix_tiles * self.batch

        # Matrices
        self.m_k_matrix = self.m * self.k
        self.k_n_matrix = self.k * self.n
        self.m_n_matrix = self.m * self.n
        self.m_k_n_matrix = self.m * self.k * self.n
        self.m_k_total_matrix = self.m_k_matrix * self.batch
        self.k_n_total_matrix = self.k_n_matrix * self.batch
        self.m_n_total_matrix = self.m_n_matrix * self.batch
        self.m_k_n_total_matrix = self.m_k_n_matrix * self.batch
        
        # Full Tiles (non-fractional tiles)
        self.m_full_tiles = math.floor(self.m / self.tile_m)
        self.k_full_tiles = math.floor(self.k / self.tile_k)
        self.n_full_tiles = math.floor(self.n / self.tile_n)

        # GEMM full tiles
        self.m_full_k_full_matrix_tiles = self.m_full_tiles * self.k_full_tiles
        self.k_full_n_full_matrix_tiles = self.k_full_tiles * self.n_full_tiles
        self.m_full_n_full_matrix_tiles = self.m_full_tiles * self.n_full_tiles
        self.m_full_k_full_n_full_matrix_tiles = self.m_full_tiles * self.k_full_tiles * self.n_full_tiles

        # GEMM full tiles across batch
        self.m_full_k_full_total_tiles = self.m_full_k_full_matrix_tiles * self.batch
        self.k_full_n_full_total_tiles = self.k_full_n_full_matrix_tiles * self.batch
        self.m_full_n_full_total_tiles = self.m_full_n_full_matrix_tiles * self.batch
        self.m_full_k_full_n_full_total_tiles = self.m_full_k_full_n_full_matrix_tiles * self.batch
        
        #Partial Tiles (fractional tiles)
        # Partial tiles within tile dimensions, can only equal 1 or 0
        self.m_partial_tiles = self.m_tiles - self.m_full_tiles
        self.k_partial_tiles = self.k_tiles - self.k_full_tiles
        self.n_partial_tiles = self.n_tiles - self.n_full_tiles
        
        # Partial tiles across tile dimension, equal to full tile dimension if partial tile exists
        self.m_full_k_partial_tiles = self.m_full_tiles * self.k_partial_tiles
        self.m_partial_k_full_tiles = self.k_full_tiles * self.m_partial_tiles
        self.k_full_n_partial_tiles = self.k_full_tiles * self.n_partial_tiles
        self.k_partial_n_full_tiles = self.n_full_tiles * self.k_partial_tiles
        self.m_full_n_partial_tiles = self.m_full_tiles * self.n_partial_tiles
        self.m_partial_n_full_tiles = self.n_full_tiles * self.m_partial_tiles

        # Partial corner tile. Only exists when both dimensions have partial tiles across tile dimensions. Can only be 1 or 0.
        self.m_partial_k_partial_tile = self.m_partial_tiles * self.k_partial_tiles
        self.k_partial_n_partial_tile = self.k_partial_tiles * self.n_partial_tiles
        self.m_partial_n_partial_tile = self.m_partial_tiles * self.n_partial_tiles

        # Partial tiles of matrix (GEMMs)
        self.m_full_k_full_n_partial_tiles = self.m_full_tiles * self.k_full_n_partial_tiles
        self.m_full_k_partial_n_full_tiles = self.m_full_k_partial_tiles * self.k_partial_n_full_tiles
        self.m_full_k_partial_n_partial_tiles = self.m_full_k_partial_tiles * self.k_partial_n_partial_tile
        self.m_partial_k_full_n_full_tiles = self.m_partial_k_full_tiles * self.n_full_tiles
        self.m_partial_k_full_n_partial_tiles = self.m_partial_k_full_tiles * self.k_full_n_partial_tiles
        self.m_partial_k_partial_n_full_tiles = self.m_partial_k_partial_tile * self.k_partial_n_full_tiles
        self.m_partial_k_partial_n_partial_tiles = self.m_partial_tiles * self.k_partial_tiles * self.n_partial_tiles

        # Partial tiles across batch 
        self.m_full_k_partial_total_tiles = self.m_full_k_partial_tiles * self.batch
        self.m_partial_k_full_total_tiles = self.m_partial_k_full_tiles * self.batch
        self.k_full_n_partial_total_tiles = self.k_full_n_partial_tiles * self.batch
        self.k_partial_n_full_total_tiles = self.k_partial_n_full_tiles * self.batch
        self.m_full_n_partial_total_tiles = self.m_full_n_partial_tiles * self.batch
        self.m_partial_n_full_total_tiles = self.m_partial_n_full_tiles * self.batch
        self.m_partial_k_partial_total_tiles = self.m_partial_k_partial_tile * self.batch
        self.k_partial_n_partial_total_tiles = self.k_partial_n_partial_tile * self.batch
        self.m_partial_n_partial_total_tiles = self.m_partial_n_partial_tile * self.batch

        # Partial tiles across batch (GEMM)
        self.m_full_k_full_n_partial_total_tiles = self.m_full_k_full_n_partial_tiles * self.batch
        self.m_full_k_partial_n_full_total_tiles = self.m_full_k_partial_n_full_tiles * self.batch
        self.m_full_k_partial_n_partial_total_tiles = self.m_full_k_partial_n_partial_tiles * self.batch
        self.m_partial_k_full_n_full_total_tiles = self.m_partial_k_full_n_full_tiles * self.batch
        self.m_partial_k_full_n_partial_total_tiles = self.m_partial_k_full_n_partial_tiles * self.batch
        self.m_partial_k_partial_n_full_total_tiles = self.m_partial_k_partial_n_full_tiles * self.batch
        self.m_partial_k_partial_n_partial_total_tiles = self.m_partial_k_partial_n_partial_tiles * self.batch

        # Partial tile sizes.
        self.tile_m_partial = self.m % self.tile_m
        self.tile_k_partial = self.k % self.tile_k
        self.tile_n_partial = self.n % self.tile_n

        #memory Size
        # Matrix memory Sizes
        self.m_k_matrix_bits = self.m_k_bitwidth * self.m * self.k
        self.k_n_matrix_bits = self.k_n_bitwidth * self.k * self.n
        self.m_n_matrix_bits = self.m_n_bitwidth * self.m * self.n
        self.m_k_total_bits = self.m_k_matrix_bits * self.batch
        self.k_n_total_bits = self.k_n_matrix_bits * self.batch
        self.m_n_total_bits = self.m_n_matrix_bits * self.batch

        # Tile memory sizes
        # Full tile memory sizes
        self.m_full_k_full_tile_bits = self.m_k_bitwidth * self.tile_m * self.tile_k
        self.k_full_n_full_tile_bits = self.k_n_bitwidth * self.tile_k * self.tile_n
        self.m_full_n_full_tile_bits = self.m_n_bitwidth * self.tile_m * self.tile_n

        # Memory Size of all full tiles
        self.m_full_k_full_matrix_bits = self.m_full_k_full_tile_bits * self.m_full_k_full_matrix_tiles
        self.k_full_n_full_matrix_bits = self.k_full_n_full_tile_bits * self.k_full_n_full_matrix_tiles
        self.m_full_n_full_matrix_bits = self.m_full_n_full_tile_bits * self.m_full_n_full_matrix_tiles

        # Memory size of all full tiles across batch
        self.m_full_k_full_total_bits = self.m_full_k_full_matrix_bits * self.batch
        self.k_full_n_full_total_bits = self.k_full_n_full_matrix_bits * self.batch
        self.m_full_n_full_total_bits = self.m_full_n_full_matrix_bits * self.batch

        # Partial tile memory sizes
        self.m_partial_k_full_tile_bits = self.m_k_bitwidth * self.tile_m_partial * self.tile_k
        self.m_full_k_partial_tile_bits = self.m_k_bitwidth * self.tile_m * self.tile_k_partial
        self.k_partial_n_full_tile_bits = self.k_n_bitwidth * self.tile_k_partial * self.tile_n
        self.k_full_n_partial_tile_bits = self.k_n_bitwidth * self.tile_k * self.tile_n_partial
        self.m_partial_n_full_tile_bits = self.m_n_bitwidth * self.tile_m_partial * self.tile_n
        self.m_full_n_partial_tile_bits = self.m_n_bitwidth * self.tile_m * self.tile_n_partial
        self.m_partial_k_partial_tile_bits = self.m_k_bitwidth * self.tile_m_partial * self.tile_k_partial
        self.k_partial_n_partial_tile_bits = self.k_n_bitwidth * self.tile_k_partial * self.tile_n_partial
        self.m_partial_n_partial_tile_bits = self.m_n_bitwidth * self.tile_m_partial * self.tile_n_partial

        # Memory size of all partial tiles
        self.m_partial_k_full_matrix_bits = self.m_partial_k_full_tile_bits * self.m_partial_k_full_tiles
        self.m_full_k_partial_matrix_bits = self.m_full_k_partial_tile_bits * self.m_full_k_partial_tiles
        self.k_partial_n_full_matrix_bits = self.k_partial_n_full_tile_bits * self.k_partial_n_full_tiles
        self.k_full_n_partial_matrix_bits = self.k_full_n_partial_tile_bits * self.k_full_n_partial_tiles
        self.m_partial_n_full_matrix_bits = self.m_partial_n_full_tile_bits * self.m_partial_n_full_tiles
        self.m_full_n_partial_matrix_bits = self.m_full_n_partial_tile_bits * self.m_full_n_partial_tiles
        self.m_partial_k_partial_matrix_bits = self.m_partial_k_partial_tile_bits * self.m_partial_k_partial_tile
        self.k_partial_n_partial_matrix_bits = self.k_partial_n_partial_tile_bits * self.k_partial_n_partial_tile
        self.m_partial_n_partial_matrix_bits = self.m_partial_n_partial_tile_bits * self.m_partial_n_partial_tile

        # Memory size of all partial tiles across batch
        self.m_partial_k_full_total_bits = self.m_partial_k_full_matrix_bits * self.batch
        self.m_full_k_partial_total_bits = self.m_full_k_partial_matrix_bits * self.batch
        self.k_partial_n_full_total_bits = self.k_partial_n_full_matrix_bits * self.batch
        self.k_full_n_partial_total_bits = self.k_full_n_partial_matrix_bits * self.batch
        self.m_partial_n_full_total_bits = self.m_partial_n_full_matrix_bits * self.batch
        self.m_full_n_partial_total_bits = self.m_full_n_partial_matrix_bits * self.batch
        self.m_partial_k_partial_total_bits = self.m_partial_k_partial_matrix_bits * self.batch
        self.k_partial_n_partial_total_bits = self.k_partial_n_partial_matrix_bits * self.batch
        self.m_partial_n_partial_total_bits = self.m_partial_n_partial_matrix_bits * self.batch
        
class TiledMatrix:
    def __init__(self, batch, m, n, tile_m, tile_n, m_n_bitwidth, array_height=None):

        if 0 in (m, n, tile_m, tile_n):
            self.is_valid = False
            return

        self.is_valid = True

        #initialize
        self.batch = batch
        self.m = m
        self.n = n
        self.tile_m = min(tile_m, self.m)
        self.tile_n = min(tile_n, self.n)
        self.m_n_bitwidth = m_n_bitwidth
        self.m_util = self.tile_m / array_height if array_height else 1

        # Batch Dims
        self.m_total = self.m * self.batch
        self.n_total = self.n * self.batch

        #Total Tiles
        self.m_tiles = math.ceil(self.m / self.tile_m)
        self.n_tiles = math.ceil(self.n / self.tile_n)
        self.m_n_matrix_tiles = self.m_tiles * self.n_tiles
        self.m_n_total_tiles = self.m_n_matrix_tiles * self.batch
    
        # Batch Tiles
        self.m_total_tiles = self.m_tiles * self.batch
        self.n_total_tiles = self.n_tiles * self.batch

        #Full Tiles (non-fractional tiles)
        self.m_full_tiles = math.floor(self.m / self.tile_m)
        self.n_full_tiles = math.floor(self.n / self.tile_n)
        self.m_full_n_full_matrix_tiles = self.m_full_tiles * self.n_full_tiles
        self.m_full_n_full_total_tiles = self.m_full_n_full_matrix_tiles * self.batch

        # Matrices
        self.m_n_matrix = self.m * self.n
        self.m_n_total_matrix = self.m_n_matrix * self.batch

        # Partial Tiles (fractional tiles)
        # Partial tiles within tile dimensions, can only equal 1 or 0
        self.m_partial_tiles = self.m_tiles - self.m_full_tiles
        self.n_partial_tiles = self.n_tiles - self.n_full_tiles
        
        # Partial tiles across tile dimension, equal to full tile dimension if partial tile exists
        self.m_full_n_partial_tiles = self.m_full_tiles * self.n_partial_tiles
        self.m_partial_n_full_tiles = self.n_full_tiles * self.m_partial_tiles

        # Partial corner tile. Only exists when both dimensions have partial tiles across tile dimensions. Can only be 1 or 0.
        self.m_partial_n_partial_tile = self.m_partial_tiles * self.n_partial_tiles

        # Partial tiles across batch 
        self.m_full_n_partial_total_tiles = self.m_full_n_partial_tiles * self.batch
        self.m_partial_n_full_total_tiles = self.m_partial_n_full_tiles * self.batch
        self.m_partial_n_partial_total_tiles = self.m_partial_n_partial_tile * self.batch

        # Partial tile sizes.
        self.tile_m_partial = self.m % self.tile_m
        self.tile_n_partial = self.n % self.tile_n

        # memory Size
        # Matrix memory Sizes
        self.m_n_matrix_bits = self.m_n_bitwidth * self.m * self.n
        self.m_n_total_bits = self.m_n_matrix_bits * self.batch

        # Tile memory sizes
        # Full tile memory sizes
        self.m_full_n_full_tile_bits = self.m_n_bitwidth * self.tile_m * self.tile_n

        # Memory Size of all full tiles
        self.m_full_n_full_matrix_bits = self.m_full_n_full_tile_bits * self.m_full_n_full_matrix_tiles

        # Memory size of all full tiles across batch
        self.m_full_n_full_total_bits = self.m_full_n_full_matrix_bits * self.batch

        # Partial tile memory sizes
        self.m_partial_n_full_tile_bits = self.m_n_bitwidth * self.tile_m_partial * self.tile_n
        self.m_full_n_partial_tile_bits = self.m_n_bitwidth * self.tile_m * self.tile_n_partial
        self.m_partial_n_partial_tile_bits = self.m_n_bitwidth * self.tile_m_partial * self.tile_n_partial

        # Memory size of all partial tiles
        self.m_partial_n_full_matrix_bits = self.m_partial_n_full_tile_bits * self.m_partial_n_full_tiles
        self.m_full_n_partial_matrix_bits = self.m_full_n_partial_tile_bits * self.m_full_n_partial_tiles
        self.m_partial_n_partial_matrix_bits = self.m_partial_n_partial_tile_bits * self.m_partial_n_partial_tile

        # Memory size of all partial tiles across batch
        self.m_partial_n_full_total_bits = self.m_partial_n_full_matrix_bits * self.batch
        self.m_full_n_partial_total_bits = self.m_full_n_partial_matrix_bits * self.batch
        self.m_partial_n_partial_total_bits = self.m_partial_n_partial_matrix_bits * self.batch


import shutil, os, sys, copy
import importlib.util

from collections import OrderedDict
from loguru import logger

from archx.utils import get_path


key_interface = 'interface'


def query_interface(module: str, query: OrderedDict, input_dir=None, output_dir=None) -> OrderedDict:
    """
    query is a dictionary with query configurations
    """
    query = copy.deepcopy(query)
    assert key_interface in query, logger.error(f'Invalid query: <{query}>. Must contain <{key_interface}>. Possible undefined attribute in archtitecture dictionary.')
    q_interface = query[key_interface]
    dst_file = os.path.join(os.path.dirname(__file__), q_interface, q_interface + '.py')

    # find proper query interface code
    spec = importlib.util.spec_from_file_location('query_' + str(module) + '_' + str(q_interface), dst_file)
    module_py = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module_py
    spec.loader.exec_module(module_py)
    
    # actual query
    del query[key_interface]
    query_result = module_py.query(module, q_interface, query, input_dir, output_dir)

    return query_result
    

def register_interface(name: str, interface_dir: str) -> None:
    assert name != key_interface, logger.error(f'Invalid interface name: <{name}>.')
    src_dir = get_path(interface_dir)
    dst_dir = os.path.join(os.path.dirname(__file__), name)
    if os.path.isdir(dst_dir):
        logger.warning(f'Interface <{name}> exists at <{dst_dir}>.')
    else:
        shutil.copytree(src_dir, dst_dir)
        logger.success(f'Register interface <{name}> from <{src_dir}> to <{dst_dir}>.')


def unregister_interface(name: str) -> None:
    assert name != key_interface, logger.error(f'Invalid interface name: <{name}>.')
    dst_dir = os.path.join(os.path.dirname(__file__), name)
    if os.path.isdir(dst_dir):
        shutil.rmtree(dst_dir)
        logger.success(f'Unregister interface <{name}> to <{dst_dir}>.')
    else:
        logger.warning(f'Interface <{name}> does not exist at <{dst_dir}>.')


def copy_interface(name: str, interface_dir: str) -> None:
    assert name != key_interface, logger.error(f'Invalid interface name: <{name}>.')
    src_dir = os.path.join(os.path.dirname(__file__), name)
    dst_dir = os.path.join(os.getcwd(), interface_dir)
    if os.path.isdir(src_dir) and not os.path.isdir(dst_dir):
        shutil.copytree(src_dir, dst_dir)
        logger.success(f'Copy interface: <{name}> from <{src_dir}> to: <{dst_dir}>.')
    elif not os.path.isdir(src_dir):
        logger.warning(f'Interface <{name}> does not exist at <{src_dir}>.')
    elif os.path.isdir(src_dir):
        logger.warning(f'Interface <{name}> exists at <{dst_dir}>.')




from typing import Callable

def condition_sweep(value, funct: Callable, condition: Callable):
    """Sweeping from a lower to upper range using a specified funct."""
    sweep_values = [value]
    new_value = funct(value)
    condition_met = condition(value)
    while condition_met:
        sweep_values.append(new_value)
        new_value = funct(new_value)
        condition_met = condition(new_value)
    return sweep_values

def iteration_sweep(value, iter: int, funct: Callable):
    """Sweeping through an explicit number of iterations using a specified funct."""
    is_single = not isinstance(value, list)
    values = [value] if is_single else list(value)
    sweep_values = []

    for _ in range(iter):
        sweep_values.append(values.copy())
        values = [funct(v) for v in values]

    if is_single:
        return [v[0] for v in sweep_values]
    return sweep_values

def list_sweep(value, funct: Callable):
    """Sweeping through a list of values using a specified funct."""
    sweep_values = []
    for v in value:
        sweep_values.append(funct(v))
    return sweep_values


from zoo.llm.common.performance.utils import TiledGEMM, TiledMatrix, sum_subevents
from archx.utils import get_prod
import math, inspect
from collections import OrderedDict

def gemm_events(tiling: list[TiledGEMM], architecture_dict: OrderedDict, workload_dict: OrderedDict, performance_dict: OrderedDict) -> OrderedDict:

    performance_dict = None

    for i, tiles in enumerate(tiling):
        if tiles.is_valid:
            if performance_dict is None:
                performance_dict = gemm_tile_events(tiles=tiles, architecture_dict=architecture_dict, workload_dict=workload_dict)
            else:
                performance_dict = sum_subevents(performance_dict, gemm_tile_events(tiles=tiles, architecture_dict=architecture_dict, workload_dict=workload_dict))

    # gemm/nonlinear mux switching event (once per gemm)
    if workload_dict['architecture'] == 'mugi':
        performance_dict['subevent']['instruction'] = OrderedDict({'count': 1})

    return performance_dict

def nonlinear_events(function: str, tiling: list[TiledMatrix], architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:

    performance_dict = None

    for tiles in tiling:
        if tiles.is_valid:
            if performance_dict is None:
                performance_dict = nonlinear_tile_events(function=function, tiles=tiles, architecture_dict=architecture_dict, workload_dict=workload_dict)
            else:
                performance_dict = sum_subevents(performance_dict, nonlinear_tile_events(function=function, tiles=tiles, architecture_dict=architecture_dict, workload_dict=workload_dict))

    # gemm/nonlinear mux switching event (once per gemm)
    if workload_dict['architecture'] == 'mugi':
        performance_dict['subevent']['instruction'] = OrderedDict({'count': 1})

    return performance_dict

def gemm_tile_events(tiles: TiledGEMM, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:

    vector_array_dim = architecture_dict['multiplier_vector']['instance'][-1]

    architecture = workload_dict['architecture']
    stationary = workload_dict['node_stationary']
    et_cycles = workload_dict.get('early_termination_cycles')
    cycles = workload_dict.get('cycles')
    
    k_dim = architecture_dict['multiplier']['instance'][-2]

    performance_dict = OrderedDict()
    performance_dict['subevent'] = OrderedDict()

    # load input
    if architecture == 'tensor':
        input_events = tiles.n_total_tiles * tiles.k_tiles if stationary == 'is' else tiles.m_n_total_tiles * tiles.k_tiles
    else:
        input_events = tiles.n_total_tiles * tiles.k if stationary == 'is' else tiles.m_n_total_tiles * tiles.k
    input_event_utilization = (tiles.m / (tiles.m_tiles * tiles.tile_m)) * tiles.m_util
    input_events *= input_event_utilization
    cycle_count_utilization = 1 / input_event_utilization
    performance_dict['subevent']['input_gemm'] = OrderedDict({'count': input_events,
                                                              'factor': {'cycle_count': cycle_count_utilization,
                                                                         'runtime': cycle_count_utilization}})
    # counter value reuse
    if architecture in ['mugi', 'carat']:
        counter_reuse_events = tiles.m_n_total_tiles * tiles.k
        counter_reuse_cycles_utalization = et_cycles / cycles if et_cycles is not None else 1
        performance_dict['subevent']['counter_reuse'] = OrderedDict({'count': counter_reuse_events,
                                                                     'factor': {'cycle_count': counter_reuse_cycles_utalization,
                                                                                'runtime': counter_reuse_cycles_utalization}})

    # broadcast value reuse
    if architecture in ['mugi', 'carat']:
        input_reuse_events = tiles.m_n_total_tiles * tiles.k
        input_reuse_event_utilization = (tiles.m / (tiles.m_tiles * tiles.tile_m)) * tiles.m_util
        input_reuse_events *= input_reuse_event_utilization
        input_reuse_cycle_utilization = et_cycles / cycles if et_cycles is not None else 1
        input_reuse_cycle_utilization *= 1 / input_reuse_event_utilization
        performance_dict['subevent']['input_reuse_gemm'] = OrderedDict({'count': input_reuse_events,
                                                                        'factor': {'cycle_count': input_reuse_cycle_utilization,
                                                                                   'runtime': input_reuse_cycle_utilization}})
    
    # load weight
    if architecture == 'tensor':
        weight_events = tiles.m_total_tiles * tiles.k_tiles if stationary == 'ws' else tiles.m_n_total_tiles * tiles.k_tiles
    else:
        weight_events = tiles.m_total_tiles * tiles.k if stationary == 'ws' else tiles.m_n_total_tiles * tiles.k
    weight_event_utilization = (tiles.n / (tiles.n_tiles * tiles.tile_n)) * tiles.n_util
    weight_events *= weight_event_utilization
    weight_cycle_utilization = 1 / weight_event_utilization
    performance_dict['subevent']['weight_gemm'] = OrderedDict({'count': weight_events,
                                                               'factor': {'cycle_count': weight_cycle_utilization,
                                                                          'runtime': weight_cycle_utilization}})

    # temporal conversion
    if architecture in ['mugi', 'carat']:
        weight_reuse_events = tiles.m_n_total_tiles * tiles.k
        weight_reuse_event_utilization = (tiles.n / (tiles.n_tiles * tiles.tile_n)) * tiles.n_util
        weight_reuse_events *= weight_reuse_event_utilization
        weight_reuse_cycle_utilization = et_cycles / cycles if et_cycles is not None else 1
        weight_reuse_cycle_utilization *= 1 / weight_reuse_event_utilization
        performance_dict['subevent']['weight_reuse_gemm'] = OrderedDict({'count': weight_reuse_events,
                                                                         'factor': {'cycle_count': weight_reuse_cycle_utilization,
                                                                                    'runtime': weight_reuse_cycle_utilization}})

    # array computation
    if architecture == 'tensor':
        array_events = tiles.m_k_n_total_tiles
    else:
        array_events = tiles.m_n_total_tiles * tiles.k
    array_events_utilization = input_event_utilization * weight_event_utilization
    array_events *= array_events_utilization
    array_cycle_utilization = et_cycles / cycles if et_cycles is not None else 1
    array_cycle_utilization *= 1 / array_events_utilization
    performance_dict['subevent']['array_gemm'] = OrderedDict({'count': array_events,
                                                              'factor': {'cycle_count': array_cycle_utilization,
                                                                         'runtime': array_cycle_utilization}})
    
    # hardwarea for multiple spikes in compute array.
    if architecture in ['mugi', 'carat']:
        array_events = tiles.m_n_total_tiles * tiles.k
        array_events_utilization = input_event_utilization * weight_event_utilization * 0.5
        array_events *= array_events_utilization
        array_cycle_utilization = et_cycles / cycles if et_cycles is not None else 1
        array_cycle_utilization *= 1 if array_events_utilization == 0 else 1 / array_events_utilization
        performance_dict['subevent']['array_fifo_gemm'] = OrderedDict({'count': array_events,
                                                                'factor': {'cycle_count': array_cycle_utilization,
                                                                            'runtime': array_cycle_utilization}})

    # vector scaling
    if architecture in ['mugi']:
        vector_events = tiles.m_total * tiles.n_tiles * (tiles.tile_n / vector_array_dim)
        vector_events_utilization = tiles.m_n_total_matrix / (vector_events * vector_array_dim)
        vector_events *= vector_events_utilization
        vector_cycle_utilization = 1 / vector_events_utilization
        performance_dict['subevent']['vector'] = OrderedDict({'count': vector_events,
                                                              'factor': {'cycle_count': vector_cycle_utilization,
                                                                         'runtime': vector_cycle_utilization}})
    elif architecture in ['carat', 'simd', 'systolic']:
        vector_events = tiles.m_total * tiles.n_tiles
        vector_events_utilization = tiles.n / (tiles.n_tiles * tiles.tile_n)
        vector_events *= vector_events_utilization
        vector_cycle_utilization = 1 / vector_events_utilization
        performance_dict['subevent']['vector_gemm'] = OrderedDict({'count': vector_events,
                                                          'factor': {'cycle_count': vector_cycle_utilization,
                                                                     'runtime': vector_cycle_utilization}})
    elif architecture in ['tensor']:
        vector_events = tiles.m_total_tiles * tiles.n_tiles
        vector_events_utilization = tiles.n / (tiles.n_tiles * tiles.tile_n)
        vector_events *= vector_events_utilization
        vector_cycle_utilization = 1 / vector_events_utilization
        performance_dict['subevent']['vector_gemm'] = OrderedDict({'count': vector_events,
                                                          'factor': {'cycle_count': vector_cycle_utilization,
                                                                     'runtime': vector_cycle_utilization}})

    # unused events
    if architecture == 'mugi':
        nonlinear_dict = 0
        performance_dict['subevent']['nonlinear_gemm'] = OrderedDict({'count': nonlinear_dict})

    return performance_dict

def nonlinear_tile_events(function: str, tiles: TiledMatrix, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:

    vector_array_dim = architecture_dict['multiplier_vector']['instance'][-1]
    architecture = workload_dict['architecture']
    performance_dict = OrderedDict()
    performance_dict['subevent'] = OrderedDict()

    # counter value reuse
    if architecture in ['mugi']:
        # input load
        input_events = tiles.m_tiles * tiles.n_total
        performance_dict['subevent']['input_nonlinear'] = OrderedDict({'count': input_events})
        
        # counter value reuse
        counter_reuse_events = tiles.m_tiles * tiles.n_total
        performance_dict['subevent']['counter_reuse'] = OrderedDict({'count': counter_reuse_events})

        # load weight (activations in nonlinear)
        weight_events = tiles.m_total_tiles * tiles.n
        weight_events_utilization = (tiles.m / (tiles.m_tiles * tiles.tile_m)) * tiles.m_util
        weight_events *= weight_events_utilization
        weight_events_cycle_utilization = 1 / weight_events_utilization if function == 'softmax' else (1 / weight_events_utilization) * 2
        performance_dict['subevent']['weight_nonlinear'] = OrderedDict({'count': weight_events,
                                                                        'factor': {'cycle_count': weight_events_cycle_utilization,
                                                                                   'runtime': weight_events_cycle_utilization}})
        
        # temporal conversion (silu maps for both signs, so mulitply by 2)
        weight_reuse_events = tiles.m_tiles * tiles.n_total
        weight_reuse_utilization = weight_events_utilization if function == 'softmax' else weight_events_utilization * 2
        weight_reuse_events *= weight_reuse_utilization
        weight_reuse_cycle_utilization = 1 / weight_reuse_utilization
        performance_dict['subevent']['weight_reuse_nonlinear'] = OrderedDict({'count': weight_reuse_events,
                                                                              'factor': {'cycle_count': weight_reuse_cycle_utilization,
                                                                                         'runtime': weight_reuse_cycle_utilization}})

        if function == 'softmax':
            # array computation (LUT selection)
            array_events = tiles.m_tiles * tiles.n_total
            array_events_utilization = weight_events_utilization
            array_events *= array_events_utilization
            array_cycle_utilization = 1 / array_events_utilization
            performance_dict['subevent']['array_nonlinear'] = OrderedDict({'count': array_events,
                                                                        'factor': {'cycle_count': array_cycle_utilization,
                                                                                    'runtime': array_cycle_utilization}})

            # summate exp
            summation_events = tiles.m_tiles * tiles.n_total
            summation_events_utilization = weight_events_utilization
            summation_events *= summation_events_utilization
            summation_cycle_utilization = 1 / summation_events_utilization
            performance_dict['subevent']['summation'] = OrderedDict({'count': summation_events,
                                                                     'factor': {'cycle_count': summation_cycle_utilization,
                                                                                'runtime': summation_cycle_utilization}})
            
            # divide by sum
            vector_events = tiles.m_tiles * (tiles.tile_m / vector_array_dim) * tiles.n_total
            vector_events_utilization = tiles.m_n_total_matrix / (vector_events * vector_array_dim)
            vector_events *= vector_events_utilization
            vector_cycle_utilization = 1 / vector_events_utilization
            performance_dict['subevent']['vector'] = OrderedDict({'count': vector_events,
                                                                  'factor': {'cycle_count': vector_cycle_utilization,
                                                                             'runtime': vector_cycle_utilization}})
        elif function == 'silu':
            # array computation (LUT selection)
            array_events = tiles.m_tiles * tiles.n_total
            array_events_utilization = weight_events_utilization * 2
            array_events *= array_events_utilization
            array_cycle_utilization = 1 / array_events_utilization
            performance_dict['subevent']['array_nonlinear'] = OrderedDict({'count': array_events,
                                                                        'factor': {'cycle_count': array_cycle_utilization,
                                                                                    'runtime': array_cycle_utilization}})

            # no sum / divide
            summation_events = 0
            performance_dict['subevent']['summation'] = OrderedDict({'count': summation_events})

            vector_events = 0
            performance_dict['subevent']['vector'] = OrderedDict({'count': vector_events})
    

        # hardware for multiple spikes in compute array.
        array_events = 0
        performance_dict['subevent']['array_fifo_nonlinear'] = OrderedDict({'count': array_events})

    # nonlinear vector
    if architecture in ['carat', 'simd', 'systolic']:
        vector_events = tiles.m_tiles * tiles.n_total
        vector_events_utilization = tiles.m / (tiles.m_tiles * tiles.tile_m) * tiles.m_util
        vector_events *= vector_events_utilization
        vector_cycle_utilization = 1 / vector_events_utilization

        if architecture in ['systolic']:
            performance_dict['subevent'][function + '_nonlinear'] = OrderedDict({'count': vector_events,
                                                                                 'factor': {'cycle_count': vector_cycle_utilization,
                                                                                            'runtime': vector_cycle_utilization}})
        else:
            performance_dict['subevent']['vector_nonlinear'] = OrderedDict({'count': vector_events,
                                                                            'factor': {'cycle_count': vector_cycle_utilization,
                                                                                       'runtime': vector_cycle_utilization}})

    if architecture in ['tensor']:
        vector_events = tiles.m_tiles * tiles.n_total
        vector_events_utilization = tiles.m / (tiles.m_tiles * tiles.tile_m) * tiles.m_util
        vector_events *= vector_events_utilization
        vector_cycle_utilization = 1 / vector_events_utilization

        performance_dict['subevent']['vector_nonlinear'] = OrderedDict({'count': vector_events,
                                                                        'factor': {'cycle_count': vector_cycle_utilization,
                                                                                    'runtime': vector_cycle_utilization}})

    # unused events
    gemm_events = 0
    performance_dict['subevent']['gemm_nonlinear'] = OrderedDict({'count': gemm_events})

    return performance_dict


from zoo.llm.results.query.utils import query_throughput_metrics, load_yaml, compute_throughput, geomean
import pandas as pd
from matplotlib import pyplot as plt
import numpy as np
from collections import OrderedDict
import os

def query(input_path, output_path):
    
    network_list = ['single_node']

    mugi_subarch_list = ['vlp', 'lut']

    vlp_list = ['mugi']
    vlp_arch_dim_list = ['128x8', '256x8']

    baseline_list = ['systolic', 'simd']
    baseline_arch_dim_list = ['8x8', '16x16']
    baseline_subarch_list = ['mac']

    throughput_module = OrderedDict({
        'mugi': OrderedDict({'gemm': 'and_gate', 'nonlinear': 'and_gate'}),
        'carat': OrderedDict({'gemm': 'and_gate', 'nonlinear': 'register_vector'}),
        'systolic': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'register_vector'}),
        'simd': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'register_vector'}),
        'tensor': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'register_vector'}),
    })

    model_list = ['llama_2_7b', 'llama_2_13b', 'llama_2_70b']
    max_seq_len_list = ['max_seq_len_128', 'max_seq_len_256', 'max_seq_len_512', 'max_seq_len_1024', 'max_seq_len_2048', 'max_seq_len_4096']
    batch_size_list = ['batch_size_1', 'batch_size_2', 'batch_size_4', 'batch_size_8', 'batch_size_16', 'batch_size_32']

    workload = 'llama_2'

    batch_size_breakdown_df = pd.DataFrame()

    for network in network_list:
        for arch in (vlp_list + baseline_list):
            for subarch in (baseline_subarch_list if arch in baseline_list else mugi_subarch_list if arch in ['mugi'] else ['']):
                subarch_label = 'none' if subarch == '' else subarch
                for arch_dim in (vlp_arch_dim_list if arch == 'mugi' else baseline_arch_dim_list if arch in baseline_list else ['8x16x16'] if arch in ['tensor'] else ['']):
                    for max_seq_len in max_seq_len_list:
                        for batch_size in batch_size_list:
                            geomean_list = []
                            for model in model_list:

                                gemm_module = throughput_module[arch]['gemm']
                                nonlinear_module = throughput_module[arch]['nonlinear']

                                termination_path = 'full_termination' if arch == 'mugi' else ''
                                run_path = os.path.normpath(f'{input_path}{arch}/{network}/{subarch}/{arch_dim}/{model}/{max_seq_len}/{batch_size}/{termination_path}/')
                                yaml_dict = load_yaml(run_path)

                                event_graph = yaml_dict['event_graph']
                                metric_dict = yaml_dict['metric_dict']

                                gemm_performance_metrics_dict = query_throughput_metrics(event_graph=event_graph, metric_dict=metric_dict, workload=workload, event='attention', module=gemm_module)

                                performance_metrics_dict = OrderedDict({
                                    'flops': gemm_performance_metrics_dict['flops'],
                                    'execution_time': gemm_performance_metrics_dict['execution_time']
                                })

                                throughput_eff_dict = compute_throughput(performance_metrics_dict=performance_metrics_dict)
                                throughput_eff_dict['energy_per_token'] = (gemm_performance_metrics_dict['energy']) / int(max_seq_len.split('_')[-1])
                                geomean_list.append(throughput_eff_dict)

                            throughput_eff_geomean_dict = geomean(geomean_list)
                            batch_size_breakdown_dict = OrderedDict({
                                'arch': arch,
                                'subarch': subarch_label,
                                'arch_dim': arch_dim,
                                'max_seq_len': max_seq_len,
                                'batch_size': batch_size,
                                'throughput': throughput_eff_geomean_dict['throughput'],
                                # 'flops': throughput_eff_geomean_dict['flops'],
                                # 'execution_time': throughput_eff_geomean_dict['execution_time'],
                            })

                            batch_size_breakdown_df = pd.concat([batch_size_breakdown_df, pd.DataFrame(batch_size_breakdown_dict, index=[0])])

    batch_size_breakdown_df.to_csv(output_path + 'gqa_size_breakdown.csv', index=False)

    baseline_df = batch_size_breakdown_df[
        (batch_size_breakdown_df['arch'] == 'systolic') &
        (batch_size_breakdown_df['subarch'] == 'mac') &
        (batch_size_breakdown_df['arch_dim'] == '8x8') &
        (batch_size_breakdown_df['batch_size'] == 'batch_size_1')
    ]

    merge_list = ['max_seq_len']
    numeric_columns = baseline_df.select_dtypes(include=['number']).columns
    columns_to_merge = merge_list + list(numeric_columns)

    merged_df = batch_size_breakdown_df.merge(
        baseline_df[columns_to_merge],
        on=merge_list,
        suffixes=('', '_baseline')
    )

    merged_df['throughput'] = merged_df['throughput'] / merged_df['throughput_baseline']
    normalized_df = merged_df.drop(
        columns=['throughput_baseline']
    )

    normalized_df.to_csv(output_path + 'batch_size_norm.csv', index=False)

def figure(input_path: str, output_path: str):
    df = pd.read_csv(input_path + 'batch_size_norm.csv')
    
    fig_width_pt = 516              
    fig_width = fig_width_pt / 72   
    fig_height = fig_width / 5.75
    fig, ax = plt.subplots(figsize=(fig_width, fig_height))

    color_map = {
        'mugi': 'dodgerblue',
        'systolic': 'grey',
        'simd': 'orangered'
    }
    
    marker_map = {
        '128x8': 'v',
        '256x8': 's',
        '8x8': 'v',
        '16x16': 's'
    }
    
    batch_sizes = [1, 2, 4, 8, 16, 32]
    seq_lens = sorted([int(seq.split('_')[-1]) for seq in df['max_seq_len'].unique()])
    segment_spacing = len(batch_sizes) + 2
    
    x_positions = []
    all_labels = []
    
    for j, seq_len in enumerate(df['max_seq_len'].unique()):
        seq_len_value = int(seq_len.split('_')[-1])
        seq_len_df = df[df['max_seq_len'] == seq_len]
        start_pos = j * segment_spacing
        x_positions_segment = np.arange(len(batch_sizes)) + start_pos
        x_positions.extend(x_positions_segment)
        
        for arch in seq_len_df['arch'].unique():
            arch_df = seq_len_df[seq_len_df['arch'] == arch]
            
            for arch_dim in arch_df['arch_dim'].unique():
                arch_dim_df = arch_df[arch_df['arch_dim'] == arch_dim]
                arch_label = 'Mugi/Carat' if arch == 'mugi' else 'SA/SA-F' if arch == 'systolic' else 'SD/SD-F'
                label = f'{arch_label} ({arch_dim.split("x")[0]})' if j == 0 else None
                if j == 0:
                    all_labels.append(label)
                color = color_map.get(arch, 'black')
                marker = marker_map.get(arch_dim, 'o')
                y_data = []

                for bs in batch_sizes:
                    batch_size = f'batch_size_{bs}'
                    value = arch_dim_df[arch_dim_df['batch_size'] == batch_size]['throughput'].values
                    y_data.append(value[0] if len(value) > 0 else np.nan)
                
                ax.plot(
                    x_positions_segment,
                    y_data,
                    label=label,
                    color=color,
                    marker=marker,
                    linestyle='-',
                    markersize=2.3,
                    linewidth=0.75,
                    alpha=0.8
                )
        
        mid_position = np.mean(x_positions_segment)
        ax.text(mid_position, ax.get_ylim()[1] + 1.75,
                f'Seq Len {seq_len_value}',
                ha='center', fontsize=5.5, fontweight='bold')
    
    ax.yaxis.set_major_formatter(plt.FuncFormatter(lambda val, pos: f'{int(val)}x'))
    ax.grid(axis='y', linestyle='--', color='grey', alpha=0.5, linewidth=0.5)  # Add gridline
    ax.set_xticks(x_positions)
    xtick_labels = [f'{bs}' for bs in batch_sizes] * len(seq_lens)
    ax.set_xticklabels(xtick_labels, fontsize=6)
    ax.set_title('Normalized Throughput', fontsize=6)
    
    ax.set_ylim(0, 40)
    ax.tick_params(axis='y', labelsize=6)
    ax.set_yticks([10, 20, 30, 40])
    
    fig.legend(
        all_labels,
        frameon=True,
        loc='upper center',
        fontsize=6,
        ncol=6,
        bbox_to_anchor=(0.5, 1.2),
    )
    
    plt.tight_layout(pad=0.2)
    plt.savefig(output_path + 'gqa_size_breakdown.pdf', dpi=1200, bbox_inches='tight')
    plt.show()


from collections import OrderedDict
from archx.utils import get_prod

def gemm(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()

    isram_bank = architecture_dict['isram']['query']['bank']
    isram_width = architecture_dict['isram']['query']['width']

    wsram_bank = architecture_dict['wsram']['query']['bank']
    wsram_width = architecture_dict['wsram']['query']['width']

    osram_bank = architecture_dict['osram']['query']['bank']
    osram_width = architecture_dict['osram']['query']['width']

    m = workload_dict['gemm']['configuration']['matrix_dim']
    k = workload_dict['gemm']['configuration']['matrix_dim']
    n = workload_dict['gemm']['configuration']['matrix_dim']
    array_dim = architecture_dict['multiplier']['instance'][-1]
    bitwidth = architecture_dict['ififo']['query']['width']

    isram_reads = (m * k * (n / array_dim) *  bitwidth) / min((isram_bank * isram_width), (array_dim * bitwidth))
    wsram_reads = (k * n * bitwidth) / min((wsram_bank * wsram_width), (array_dim * bitwidth))
    osram_writes = (m * k * (n / array_dim) * bitwidth) / min((osram_bank * osram_width), (array_dim * bitwidth))

    input_mappings = (m / array_dim) * (k / array_dim) * (n)
    weight_mappings = (m / array_dim) * (k / array_dim) * (n)
    # weight_mappings = (k / array_dim) * (n)

    isram_reads_dict = OrderedDict({'count': isram_reads})
    wsram_reads_dict = OrderedDict({'count': wsram_reads})
    osram_writes_dict = OrderedDict({'count': osram_writes})
    input_array_dict = OrderedDict({'count': input_mappings})
    weight_array_dict = OrderedDict({'count': weight_mappings})

    performance_dict['subevent'] = OrderedDict({'input_reads': isram_reads_dict,
                                                'weight_reads': wsram_reads_dict,
                                                'output_writes': osram_writes_dict,
                                                'input_array': input_array_dict,
                                                'weight_array': weight_array_dict})

    return performance_dict

def input_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    frequency = architecture_dict['isram']['query']['frequency']

    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    performance_dict['cycle_count'] = OrderedDict({'value': 1, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / frequency, 'unit': 'ms'})
    sram_dict = OrderedDict({'operation': OrderedDict({'dynamic_energy': 'read'})})
    performance_dict['subevent'] = OrderedDict({'isram': sram_dict})

    return performance_dict

def weight_reads(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    frequency = architecture_dict['wsram']['query']['frequency']

    performance_dict['cycle_count'] = OrderedDict({'value': 1, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / frequency, 'unit': 'ms'})
    sram_dict = OrderedDict({'operation': OrderedDict({'dynamic_energy': 'read'})})
    performance_dict['subevent'] = OrderedDict({'wsram': sram_dict})

    return performance_dict

def output_writes(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    frequency = architecture_dict['osram']['query']['frequency']

    performance_dict['cycle_count'] = OrderedDict({'value': 1, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / frequency, 'unit': 'ms'})
    sram_dict = OrderedDict({'operation': OrderedDict({'dynamic_energy': 'write'})})
    performance_dict['subevent'] = OrderedDict({'osram': sram_dict})

    return performance_dict

def input_array(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()

    ififo_dim = get_prod(architecture_dict['ififo']['instance'])
    wfifo_dim = get_prod(architecture_dict['wfifo']['instance'])
    ofifo_dim = get_prod(architecture_dict['ofifo']['instance'])
    output_adder_dim = get_prod(architecture_dict['output_adder']['instance'])
    multiplier_dim = get_prod(architecture_dict['multiplier']['instance'])
    adder_dim = get_prod(architecture_dict['adder']['instance'])
    act_en_reg_dim = get_prod(architecture_dict['act_en_reg']['instance'])
    mult_en_reg_dim = get_prod(architecture_dict['mult_en_reg']['instance'])
    acc_en_reg_dim = get_prod(architecture_dict['acc_en_reg']['instance'])
    weight_path_en_reg_dim = get_prod(architecture_dict['weight_path_en_reg']['instance'])
    weight_en_reg_dim = get_prod(architecture_dict['weight_en_reg']['instance'])
    sum_en_reg_dim = get_prod(architecture_dict['sum_en_reg']['instance'])
    act_reg_dim = get_prod(architecture_dict['act_reg']['instance'])
    weight_path_reg_dim = get_prod(architecture_dict['weight_path_reg']['instance'])
    sum_reg_dim = get_prod(architecture_dict['sum_reg']['instance'])
    weight_reg_dim = get_prod(architecture_dict['weight_reg']['instance'])
    act_mux_dim = get_prod(architecture_dict['act_mux']['instance'])
    weight_mux_dim = get_prod(architecture_dict['weight_mux']['instance'])
    add_mux_dim = get_prod(architecture_dict['add_mux']['instance'])
    sum_mux_dim = get_prod(architecture_dict['sum_mux']['instance'])

    frequency = architecture_dict['multiplier']['query']['frequency']
    performance_dict['cycle_count'] = OrderedDict({'value': 1, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / frequency, 'unit': 'ms'})

    ififo_dict = OrderedDict({'count': ififo_dim})
    wfifo_dict = OrderedDict({'count': wfifo_dim})
    ofifo_dict = OrderedDict({'count': ofifo_dim})
    output_adder_dict = OrderedDict({'count': output_adder_dim})
    multiplier_dict = OrderedDict({'count': multiplier_dim})
    adder_dict = OrderedDict({'count': adder_dim})
    act_en_reg_dict = OrderedDict({'count': act_en_reg_dim})
    mult_en_reg_dict = OrderedDict({'count': mult_en_reg_dim})
    acc_en_reg_dict = OrderedDict({'count': acc_en_reg_dim})
    weight_path_en_reg_dict = OrderedDict({'count': weight_path_en_reg_dim})
    weight_en_reg_dict = OrderedDict({'count': weight_en_reg_dim})
    sum_en_reg_dict = OrderedDict({'count': sum_en_reg_dim})
    act_reg_dict = OrderedDict({'count': act_reg_dim})
    weight_path_reg_dict = OrderedDict({'count': weight_path_reg_dim})
    sum_reg_dict = OrderedDict({'count': sum_reg_dim})
    weight_reg_dict = OrderedDict({'count': weight_reg_dim})
    act_mux_dict = OrderedDict({'count': act_mux_dim})
    weight_mux_dict = OrderedDict({'count': weight_mux_dim})
    add_mux_dict = OrderedDict({'count': add_mux_dim})
    sum_mux_dict = OrderedDict({'count': sum_mux_dim})

    performance_dict['subevent'] = OrderedDict({'ififo': ififo_dict,
                                                'wfifo': wfifo_dict,
                                                'ofifo': ofifo_dict,
                                                'output_adder': output_adder_dict,
                                                'multiplier': multiplier_dict,
                                                'adder': adder_dict,
                                                'act_en_reg': act_en_reg_dict,
                                                'mult_en_reg': mult_en_reg_dict,
                                                'acc_en_reg': acc_en_reg_dict,
                                                'weight_path_en_reg': weight_path_en_reg_dict,
                                                'weight_en_reg': weight_en_reg_dict,
                                                'sum_en_reg': sum_en_reg_dict,
                                                'act_reg': act_reg_dict,
                                                'weight_path_reg': weight_path_reg_dict,
                                                'sum_reg': sum_reg_dict,
                                                'weight_reg': weight_reg_dict,
                                                'act_mux': act_mux_dict,
                                                'weight_mux': weight_mux_dict,
                                                'add_mux': add_mux_dict,
                                                'sum_mux': sum_mux_dict})
    return performance_dict

def weight_array(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()

    frequency = architecture_dict['multiplier']['query']['frequency']
    performance_dict['cycle_count'] = OrderedDict({'value': 1, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / frequency, 'unit': 'ms'})

    weight_path_en_reg_dim = get_prod(architecture_dict['weight_path_en_reg']['instance'])
    weight_en_reg_dim = get_prod(architecture_dict['weight_en_reg']['instance'])
    weight_path_reg_dim = get_prod(architecture_dict['weight_path_reg']['instance'])
    weight_reg_dim = get_prod(architecture_dict['weight_reg']['instance'])

    weight_path_en_reg_dict = OrderedDict({'count': weight_path_en_reg_dim})
    weight_en_reg_dict = OrderedDict({'count': weight_en_reg_dim})
    weight_path_reg_dict = OrderedDict({'count': weight_path_reg_dim})
    weight_reg_dict = OrderedDict({'count': weight_reg_dim})

    performance_dict['subevent'] = OrderedDict({'weight_path_en_reg': weight_path_en_reg_dict,
                                                'weight_en_reg': weight_en_reg_dict,
                                                'weight_path_reg': weight_path_reg_dict,
                                                'weight_reg': weight_reg_dict})
    return performance_dict


import pyfiglet, argparse, time, os, sys
import pandas as pd
import subprocess
import shutil

from loguru import logger

from archx.architecture import create_architecture_dict, save_architecture_dict
from archx.event import create_event_graph, save_event_graph
from archx.metric import create_metric_dict, create_event_metrics, save_metric_dict
from archx.workload import create_workload_dict, save_workload_dict
from archx.performance import simulate_performance_all_events
from archx.utils import bcolors, write_yaml, read_yaml
from archx.interface import register_interface, unregister_interface, copy_interface
from archx.programming.graph.agraph import AGraph, _generate_runs, _gui

def parse_commandline_args():
    """
    parse command line inputs
    """
    parser = argparse.ArgumentParser(
        description='Archx is a framework to explore the design space of hardware architecture.')
    parser.add_argument('-r', '--run_dir', type=str, default=None,
                        help = 'Run directory to store outputs.')
    parser.add_argument('-a', '--architecture_yaml', type=str, default=None,
                        help = 'Path to architecture yaml.')
    parser.add_argument('-m', '--metric_yaml', type=str, default=None,
                        help = 'Path to metric yaml.')
    parser.add_argument('-w', '--workload_yaml', type=str, default=None,
                        help = 'Path to workload yaml.')
    parser.add_argument('-e', '--event_yaml', type=str, default=None,
                        help = 'Path to event yaml.')
    parser.add_argument('-c', '--checkpoint', type=str, default=None,
                        help = 'Path to checkpoint, which requires <.gt> format.')
    parser.add_argument('-l', '--log_level', type=str, default='INFO',
                        help = 'Level of logger.')
    parser.add_argument('-d', '--delete', action='store_true', default=False,
                        help = 'Delete run directory if it exists.')
    parser.add_argument('-s', '--save_yaml', action='store_true', default=False, help = 'Save yaml files in run directory.')
    parser.add_argument('-ireg', '--register_interface', action='store_true', default=False, help = 'Register a new interface.')
    parser.add_argument('-iureg', '--unregister_interface', action='store_true', default=False, help = 'Unregister a new interface.')
    parser.add_argument('-icopy', '--copy_interface', action='store_true', default=False, help = 'Copy an existing interface.')
    parser.add_argument('-iname', '--interface_name', type=str, default=None, help = 'Name of the interface.')
    parser.add_argument('-idir', '--interface_dir', type=str, default=None, help = 'Directory of the interface.')

    # frontend programming parse
    parser.add_argument('-compile', '--compile', type=str, default=None,
                        help = 'Path to archx frontent python file for configuration sweeping.')
    parser.add_argument('-extract', '--extract', type=str, default=None,
                        help = 'Extract unfiltered configuration results from generated csv file.')
    parser.add_argument('-f', '--filter', type=str, default=None,
                        help = 'Open GUI to filter configurations from generated csv file.')
    parser.add_argument('-ff', '--filter_full', action='store_true', default=False,
                        help = 'Enable full mode for frontend programming filtering (filter + execute).')
    parser.add_argument('-t', '--tabular', action='store_true', default=False,
                        help = 'Add logger debug to terminal output.')
    parser.add_argument('-x', '--execute', type=str, default=None,
                        help = 'Path to generated configuration runs text file for execution.')
    parser.add_argument('-full', '--full', action='store_true', default=False,
                        help = 'Enable full mode for frontend programming (compile, extract, filter, execute).')

    return parser.parse_args()


def main():
    args = parse_commandline_args()
    
    # set up banner
    ascii_banner = pyfiglet.figlet_format('Archx')
    print(bcolors.Magenta + ascii_banner + bcolors.ENDC)
    ascii_banner = pyfiglet.figlet_format('UnaryLab')
    print(bcolors.Yellow + ascii_banner + bcolors.ENDC)
    ascii_banner = pyfiglet.figlet_format('https://github.com/UnaryLab/archx', font='term')
    print(bcolors.UNDERLINE + bcolors.Green + ascii_banner + bcolors.ENDC)

    # check not all interface options are selected
    assert (args.register_interface + args.unregister_interface + args.copy_interface <= 1), logger.error('Only one interface option can be selected at a time: <-regi>, <-uregi>, or <-copyi>.')
    
    if args.delete:
        # delete run directory if it exists
        if os.path.isdir(args.run_dir):
            shutil.rmtree(args.run_dir)
            logger.info(f'Delete existing run directory <{args.run_dir}>.')

    # register interface
    if args.register_interface:
        assert args.interface_name is not None, logger.error(f'Interface name is required for registration via <-iname>.')
        assert args.interface_dir is not None, logger.error(f'Interface directory is required for registration via <-idir>.')
        register_interface(args.interface_name, args.interface_dir)
        exit(0)
    # unregister interface
    elif args.unregister_interface:
        assert args.interface_name is not None, logger.error(f'Interface name is required for unregistration via <-iname>.')
        unregister_interface(args.interface_name)
        exit(0)
    # copy interface
    elif args.copy_interface:
        assert args.interface_name is not None, logger.error(f'Interface name is required for copying via <-iname>.')
        assert args.interface_dir is not None, logger.error(f'Interface directory is required for copying via <-idir>.')
        copy_interface(args.interface_name, args.interface_dir)
        exit(0)


    # validate run dir exists and path is valid
    assert args.run_dir is not None, logger.error(f'Run directory is required via <-r> or <-run_dir>.')
    # check if run directory exists, if not create it
    if not os.path.isdir(args.run_dir):
        os.makedirs(args.run_dir)
        logger.info(f'Create run directory <{args.run_dir}>.')
    else:
        logger.info(f'Find run directory <{args.run_dir}>.')
    if not args.run_dir.endswith('/'):
        args.run_dir += '/'

    # set up output log
    logger.remove()
    output_log = args.run_dir + '/archx' + '-' + str(time.time()) + '.log'
    logger.add(output_log, level=args.log_level)

    if args.tabular:
        logger.add(sys.stdout, level=args.log_level)
        logger.info('Enable logger output to terminal.')

    archx_run = args.architecture_yaml is not None and args.metric_yaml is not None and args.workload_yaml is not None and args.event_yaml is not None and args.checkpoint is not None

    tabular_mode = "-t" if args.tabular else ""

    if archx_run:
        # validate checkpoint
        assert args.checkpoint.endswith('.gt'), logger.error('Invalid event checkpoint format; requires <.gt>.')

        logger.success(f'\n----------------------------------------------\nStep 1: Create architectue dict\n----------------------------------------------')
        architecture_dict = create_architecture_dict(args.architecture_yaml)

        logger.success(f'\n----------------------------------------------\nStep 2: Create metric dict\n----------------------------------------------')
        metric_dict = create_metric_dict(args.metric_yaml)

        logger.success(f'\n----------------------------------------------\nStep 3: Creat workload dict\n----------------------------------------------')
        workload_dict = create_workload_dict(args.workload_yaml)
        
        logger.success(f'\n----------------------------------------------\nStep 4: Creat event graph\n----------------------------------------------')
        event_graph = create_event_graph(args.event_yaml)

        logger.success(f'\n----------------------------------------------\nStep 5: Create metrics for all events and modules\n----------------------------------------------')
        event_graph = create_event_metrics(event_graph, architecture_dict, metric_dict, run_dir=args.run_dir)
        
        logger.success(f'\n----------------------------------------------\nStep 6: Simulate performance\n----------------------------------------------')
        event_graph = simulate_performance_all_events(event_graph, architecture_dict, workload_dict)
        
        logger.success(f'\n----------------------------------------------\nStep 7: Save event graph and log\n----------------------------------------------')
        save_event_graph(event_graph=event_graph, save_path=args.checkpoint)

        if args.save_yaml:
            save_architecture_dict(architecture_dict=architecture_dict, save_path=args.run_dir + '/architecture.yaml')
            save_metric_dict(metric_dict=metric_dict, save_path=args.run_dir + '/metric.yaml')
            save_workload_dict(workload_dict=workload_dict, save_path=args.run_dir + '/workload.yaml')
            write_yaml(args.run_dir + '/event.yaml', read_yaml(args.event_yaml))
            logger.success(f'Save dictionaries to <{args.run_dir}>.')

        logger.success(f'Save log to <{output_log}>.')
    elif args.compile:
        # frontend programming compile mode
        compile_path = args.compile
        assert os.path.isfile(compile_path), logger.error(f'Invalid compile file path <{compile_path}>.')

        # import compile file as module
        import importlib.util
        spec = importlib.util.spec_from_file_location("compile_module", compile_path)
        compile_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(compile_module) 
        # call compile function
        assert hasattr(compile_module, 'description'), logger.error(f'Compile file <{compile_path}> does not contain <description()> function.')
        agraph = compile_module.description(path=args.run_dir)
        logger.success(f'Compile generation completed in <{args.run_dir}>.')
        
        if args.full:
            extract_path = args.run_dir + '/configurations.csv'
            df = pd.read_csv(extract_path)
            _generate_runs(df=df, path=args.run_dir + '/runs.txt')

            if args.filter_full:
                filter_path = args.run_dir + '/configurations.csv'
                df = pd.read_csv(filter_path)
                _gui(df=df, path=args.run_dir + '/runs.txt')
            else:
                execute_path = args.run_dir + '/runs.txt'
                assert os.path.isfile(execute_path), logger.error(f'Invalid execute runs file <{execute_path}>.')

            # call run_archx.sh with runs file
            script_path = os.path.join(os.path.dirname(__file__), 'bin', 'run_archx.sh')
            command = f'bash {script_path} {execute_path} {tabular_mode}'
            logger.info(f'Executing command: {command}')
            process = subprocess.Popen(command, shell=True)
            process.communicate()

            logger.success(f'Execution of runs in <{execute_path}> completed.')

        
    elif args.extract:
        # frontend programming extract mode
        extract_path = args.extract
        assert os.path.isfile(extract_path), logger.error(f'Invalid extract csv file <{extract_path}>.')

        df = pd.read_csv(extract_path)
        _generate_runs(df=df, path=args.run_dir + '/runs.txt')
        
        logger.success(f'Extract runs.txt file compiled in <{extract_path}>.')

    elif args.filter:
        # frontend programming filter mode
        filter_path = args.filter
        assert os.path.isfile(filter_path), logger.error(f'Invalid filter csv file <{filter_path}>.')

        df = pd.read_csv(filter_path)
        _gui(df=df, path=args.run_dir + '/runs.txt')
        
        logger.success(f'Filter runs.txt file compiled in <{filter_path}>.')

    elif args.execute:
        # frontend programming execute mode
        execute_path = args.execute
        assert os.path.isfile(execute_path), logger.error(f'Invalid execute runs file <{execute_path}>.')

        # call run_archx.sh with runs file
        script_path = os.path.join(os.path.dirname(__file__), 'bin', 'run_archx.sh')
        command = f'bash {script_path} {execute_path} {tabular_mode}'
        logger.info(f'Executing command: {command}')
        process = subprocess.Popen(command, shell=True)
        process.communicate()

        logger.success(f'Execution of runs in <{execute_path}> completed.')


if __name__ == '__main__':
    main()




from zoo.llm.common.performance.utils import TiledGEMM, TiledMatrix
import math
from collections import OrderedDict

def router_gemm_events(tiles: TiledGEMM, offchip_memory_events_dict: OrderedDict, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:
    """
    Estimates events from router to onchip memory for GEMM operations.
    Needs dram <-> sram performance events passed.
    """
    performance_dict = OrderedDict()

    # Retrieve dicts
    router_height = architecture_dict['irouter']['instance'][0]
    router_width = architecture_dict['irouter']['instance'][1]
    router_dim = router_height * router_width
    scheduling = workload_dict['noc_stationary']
    osram_width = architecture_dict['osram']['query']['width']
    isram_width = architecture_dict['isram']['query']['width']
    wsram_width = architecture_dict['wsram']['query']['width']

    # adjust based on scheduling
    if scheduling == 'os':
        m_k_tiles = tiles.m_k_n_total_tiles
        k_n_tiles = tiles.m_k_n_total_tiles
        m_n_tiles = tiles.m_n_total_tiles
    elif scheduling == 'is':
        m_k_tiles = tiles.m_k_total_tiles
        k_n_tiles = tiles.m_k_n_total_tiles
        m_n_tiles = tiles.m_k_n_total_tiles
    elif scheduling == 'ws':
        m_k_tiles = tiles.m_k_n_total_tiles
        k_n_tiles = tiles.k_n_total_tiles
        m_n_tiles = tiles.m_k_n_total_tiles

    # calculate average router event across all mappings
    m_k_full_router_events = math.floor(m_k_tiles / router_dim)
    k_n_full_router_events = math.floor(k_n_tiles / router_dim)
    m_n_full_router_events = math.floor(m_n_tiles / router_dim)

    m_k_partial_router_events = 1 if m_k_tiles % router_dim > 0 else 0
    k_n_partial_router_events = 1 if k_n_tiles % router_dim > 0 else 0
    m_n_partial_router_events = 1 if m_n_tiles % router_dim > 0 else 0

    m_k_mappings = m_k_full_router_events + m_k_partial_router_events
    k_n_mappings = k_n_full_router_events + k_n_partial_router_events
    m_n_mappings = m_n_full_router_events + m_n_partial_router_events

    m_k_partial_router_height = round(math.sqrt(m_k_partial_router_events))
    k_n_partial_router_height = round(math.sqrt(k_n_partial_router_events))
    m_n_partial_router_height = round(math.sqrt(m_n_partial_router_events))

    m_k_average_router_events = ((router_height * m_k_full_router_events) + (m_k_partial_router_height * m_k_partial_router_events)) / m_k_mappings
    k_n_average_router_events = ((router_height * k_n_full_router_events) + (k_n_partial_router_height * k_n_partial_router_events)) / k_n_mappings
    m_n_average_router_events = ((router_height * m_n_full_router_events) + (m_n_partial_router_height * m_n_partial_router_events)) / m_n_mappings

    irouter_count = m_k_average_router_events * offchip_memory_events_dict['subevent']['isram_offchip_writes']['count'] * osram_width
    wrouter_count = k_n_average_router_events * offchip_memory_events_dict['subevent']['wsram_offchip_writes']['count'] * isram_width
    orouter_count = m_n_average_router_events * offchip_memory_events_dict['subevent']['osram_offchip_reads']['count'] * wsram_width

    irouter_dict = OrderedDict({'count': irouter_count})
    wrouter_dict = OrderedDict({'count': wrouter_count})
    orouter_dict = OrderedDict({'count': orouter_count})

    performance_dict['subevent'] = OrderedDict({
        'irouter_mapping': irouter_dict,
        'wrouter_mapping': wrouter_dict,
        'orouter_mapping': orouter_dict
    })

    return performance_dict

def router_nonlinear_events(tiles: TiledMatrix, offchip_memory_events_dict: OrderedDict, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> OrderedDict:
    """
    Estimates events from router to onchip memory for GEMM operations.
    Needs dram <-> sram performance events passed within performance_dict.
    """
    offchip_memory_events_dict = offchip_memory_events_dict['subevent']
    performance_dict = OrderedDict()

    # Retrieve dicts
    router_height = architecture_dict['irouter']['instance'][0]
    router_width = architecture_dict['irouter']['instance'][1]
    router_dim = router_height * router_width
    architecture = workload_dict['architecture']

    # calculate average router event across all mappings
    m_n_full_router_events = math.floor(tiles.m_n_total_tiles / router_dim)
    m_n_partial_router_events = 1 if tiles.m_n_total_tiles % router_dim > 0 else 0
    m_n_mappings = m_n_full_router_events + m_n_partial_router_events

    m_n_partial_router_height = round(math.sqrt(m_n_partial_router_events))
    m_n_average_router_events = ((router_height * m_n_full_router_events) + (m_n_partial_router_height * m_n_partial_router_events)) / m_n_mappings

    irouter_average_events = m_n_average_router_events if tiles.m_n_total_tiles < router_dim else router_height

    if architecture == 'mugi':
        irouter_count = irouter_average_events * offchip_memory_events_dict['isram_offchip_writes']['count']
        wrouter_count = m_n_average_router_events * offchip_memory_events_dict['wsram_offchip_writes']['count']
        orouter_count = m_n_average_router_events * (offchip_memory_events_dict['osram_offchip_reads']['count'] + offchip_memory_events_dict['osram_offchip_writes']['count'])
    else:
        irouter_count = 0
        wrouter_count = 0
        orouter_count = m_n_average_router_events * (offchip_memory_events_dict['osram_offchip_reads']['count'] + offchip_memory_events_dict['osram_offchip_writes']['count'])

    irouter_dict = OrderedDict({'count': irouter_count})
    wrouter_dict = OrderedDict({'count': wrouter_count})
    orouter_dict = OrderedDict({'count': orouter_count})

    performance_dict['subevent'] = OrderedDict({
        'irouter_mapping': irouter_dict,
        'wrouter_mapping': wrouter_dict,
        'orouter_mapping': orouter_dict
    })

    return performance_dict



from loguru import logger
from copy import deepcopy

class Workload():
    def __init__(self, parameter_enumerator):
        self.parameter_enumerator = parameter_enumerator
        self.configurations = []

    def add_configuration(self, name: str):
        assert isinstance(name, str), "'name' parameter must be of type 'str'."
        assert all(config.name != name for config in self.configurations), f"Configuration '{name}' already exists in workload."

        configuration = Configuration(name, self.parameter_enumerator)
        self.configurations.append(configuration)
        logger.info(f"Added configuration '{name}'.")
        return configuration
    
    def to_yaml(self, config):
        workload_dict = {'workload': {}}

        for var_name, value in config.items():
            param_info = self.parameter_enumerator.get_parameters_from_name(var_name)
            config_name = param_info['name']
            param_name = param_info['param_name']
            copy_value = deepcopy(value)
            
            if config_name not in workload_dict['workload']:
                workload_dict['workload'][config_name] = {'configuration': {}}
            workload_dict['workload'][config_name]['configuration'][param_name] = copy_value
        
        return workload_dict
            

class Configuration():
    def __init__(self, name: str, parameter_enumerator):
        self.name = name
        self.parameter_enumerator = parameter_enumerator

    def add_parameter(self, parameter_name: str, parameter_value, sweep: bool = None):
        assert isinstance(parameter_name, str), "'parameter_name' parameter must be of type 'str'."
        assert isinstance(sweep, (bool, type(None))), "'sweep' parameter must be of type 'bool'."
        assert (isinstance(parameter_value, list) and isinstance(sweep, bool)) or (not isinstance(parameter_value, list)) , "If 'parameter_value' is a list, 'sweep' must be defined as either True or False."

        sweep = sweep if sweep is not None else False

        parameter_param = self.parameter_enumerator.add_parameter(name=self.name,
                                                                   param_name=parameter_name,
                                                                   value=parameter_value,
                                                                   type='parameter',
                                                                   sweep=sweep,
                                                                   desc='workload')

        param_dict = {
            parameter_name: parameter_param
        }

        logger.info(f"Added parameter '{parameter_name}' to configuration '{self.name}'.")
        logger.debug(f"\tValue: {parameter_value}")
        logger.debug(f"\tSweep: {sweep}")

        return param_dict


from ortools.sat.python import cp_model

class ParameterEnumerator:
    def __init__(self, model: cp_model.CpModel):
        self.model = model
        self.parameters = {}  # IntVar -> {desc, type, name, param_name}
        self.var_to_values = {}  # IntVar -> values list (for fast lookup)
        self.name_to_var = {}  # var name string -> IntVar (for reverse lookup)
    
    def add_parameter(self, name: str, param_name: str, value, type: str, sweep: bool, desc: str):
        """
        Add a parameter with multiple possible values.
        
        Args:
            name: Unique identifier for this parameter
            values: List of possible values (e.g., [32, 64, 128] or ['small', 'large'])
        
        Returns:
            OR-Tools IntVar representing index into values list
        """

        full_name = f"{desc}_{type}_{name}_{param_name}"

        if sweep:
            # Multiple values - domain is [0, len-1]
            var = self.model.NewIntVar(0, len(value) - 1, full_name)
            
        else:
            # Single value - still create variable but with domain [0, 0]
            var = self.model.NewIntVar(0, 0, full_name)

        values_list = value if sweep else [value]
            
        self.parameters[var] = {
            'desc': desc,
            'type': type,
            'name': name,
            'param_name': param_name
        }

        # Store flat mapping for fast lookup by var
        self.var_to_values[var] = values_list
        
        # Store reverse lookup by name string
        self.name_to_var[full_name] = var
        
        return var
    
    def get_parameters_from_var(self, var: cp_model.IntVar):
        """Retrieve parameter info given an IntVar."""
        assert isinstance(var, cp_model.IntVar), "'var' must be an instance of cp_model.IntVar."
        assert var in self.parameters, f"IntVar '{var.Name()}' not found in parameter enumerator."
        return self.parameters[var]
    
    def get_parameters_from_name(self, var_name: str):
        """Retrieve parameter info given a variable name string."""
        assert isinstance(var_name, str), "'var_name' must be a string."
        assert var_name in self.name_to_var, f"Variable '{var_name}' not found in parameter enumerator."
        var = self.name_to_var[var_name]
        return self.parameters[var]
    
    def get_values_for_var(self, var: cp_model.IntVar):
        """Retrieve the list of actual values for a given IntVar."""
        assert isinstance(var, cp_model.IntVar), "'var' must be an instance of cp_model.IntVar."
        assert var in self.var_to_values, f"IntVar '{var.Name()}' not found in parameter enumerator."
        return self.var_to_values[var]


from collections import OrderedDict
from archx.utils import get_prod

def input_gemm(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    frequency = architecture_dict['ififo']['query']['frequency']
    ififo_dim = get_prod(architecture_dict['ififo']['instance'][-2:])
    input_register_dim = get_prod(architecture_dict['input_register']['instance'][-3:])

    cycle_count = 1 / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})
    
    ififo_dict = OrderedDict({'count': ififo_dim})
    input_register_dict = OrderedDict({'count': input_register_dim})

    performance_dict['subevent'] = OrderedDict({'ififo': ififo_dict,
                                                'input_register': input_register_dict})
    return performance_dict

def weight_gemm(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    frequency = architecture_dict['wfifo']['query']['frequency']
    wfifo_dim = get_prod(architecture_dict['wfifo']['instance'][-2:])
    weight_register_dim = get_prod(architecture_dict['weight_register']['instance'][-3:])

    if 'int_to_fp' in architecture_dict:
        int_to_fp_dim = get_prod(architecture_dict['int_to_fp']['instance'][-2:])

    cycle_count = 1 / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})
    
    wfifo_dict = OrderedDict({'count': wfifo_dim})
    weight_register_dict = OrderedDict({'count': weight_register_dim})
    if 'int_to_fp' in architecture_dict:
        int_to_fp_dict = OrderedDict({'count': int_to_fp_dim})

    performance_dict['subevent'] = OrderedDict({'wfifo': wfifo_dict,
                                                'weight_register': weight_register_dict})
    if 'int_to_fp' in architecture_dict:
        performance_dict['subevent']['int_to_fp'] = int_to_fp_dict
    return performance_dict

def array_gemm(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    frequency = architecture_dict['multiplier']['query']['frequency']

    multiplier_dim = get_prod(architecture_dict['multiplier']['instance'][-3:])
    pe_register_dim = get_prod(architecture_dict['pe_register']['instance'][-3:])
    accumulator_dim = get_prod(architecture_dict['accumulator']['instance'][-3:])
    accumulator_register_dim = get_prod(architecture_dict['accumulator_register']['instance'][-2:])
    adder_dim = get_prod(architecture_dict['adder']['instance'][-2:])
    ofifo_dim = get_prod(architecture_dict['ofifo']['instance'][-2:])

    if 'ch_aloc' in architecture_dict:
        ch_aloc_dim = get_prod(architecture_dict['ch_aloc']['instance'][-2:])
    if 'ch_dealoc' in architecture_dict:
        ch_dealoc_dim = get_prod(architecture_dict['ch_dealoc']['instance'][-2:])
    if 'int_to_fp_figna' in architecture_dict:
        int_to_fp_figna_dim = get_prod(architecture_dict['int_to_fp_figna']['instance'][-2:])
    if 'prealigner' in architecture_dict:
        prealigner_dim = get_prod(architecture_dict['prealigner']['instance'][-2:])

    cycle_count = 1 / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})
    
    multiplier_dict = OrderedDict({'count': multiplier_dim})
    pe_register_dict = OrderedDict({'count': pe_register_dim})
    accumulator_dict = OrderedDict({'count': accumulator_dim})
    accumulator_register_dict = OrderedDict({'count': accumulator_register_dim * cycle_count})
    adder_dict = OrderedDict({'count': adder_dim * cycle_count})
    ofifo_dict = OrderedDict({'count': ofifo_dim * cycle_count})

    if 'ch_aloc' in architecture_dict:
        ch_aloc_dict = OrderedDict({'count': ch_aloc_dim})
    if 'ch_dealoc' in architecture_dict:
        ch_dealoc_dict = OrderedDict({'count': ch_dealoc_dim})
    if 'int_to_fp_figna' in architecture_dict:
        int_to_fp_figna_dict = OrderedDict({'count': int_to_fp_figna_dim})
    if 'prealigner' in architecture_dict:
        prealigner_dict = OrderedDict({'count': prealigner_dim})

    performance_dict['subevent'] = OrderedDict({'multiplier': multiplier_dict,
                                                'pe_register': pe_register_dict,
                                                'accumulator': accumulator_dict,
                                                'accumulator_register': accumulator_register_dict,
                                                'adder': adder_dict,
                                                'ofifo': ofifo_dict})
    if 'ch_aloc' in architecture_dict:
        performance_dict['subevent']['ch_aloc'] = ch_aloc_dict
    if 'ch_dealoc' in architecture_dict:
        performance_dict['subevent']['ch_dealoc'] = ch_dealoc_dict
    if 'int_to_fp_figna' in architecture_dict:
        performance_dict['subevent']['int_to_fp_figna'] = int_to_fp_figna_dict
    if 'prealigner' in architecture_dict:
        performance_dict['subevent']['prealigner'] = prealigner_dict

    return performance_dict

def vector_gemm(architecture_dict: OrderedDict, workload_dict: OrderedDict=None)->OrderedDict:
    performance_dict = OrderedDict()
    router_dim = get_prod(architecture_dict['irouter']['instance']) if 'irouter' in architecture_dict else 1

    frequency = architecture_dict['multiplier_vector']['query']['frequency']
    multiplier_vector_dim = get_prod(architecture_dict['multiplier_vector']['instance'][-2:])
    register_vector_dim = get_prod(architecture_dict['register_vector']['instance'][-2:])

    cycle_count = 1 / router_dim
    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})
    
    multiplier_vector_dict = OrderedDict({'count': multiplier_vector_dim})
    accumulator_vector_dict = OrderedDict({'count': 0})
    register_vector_dict = OrderedDict({'count': register_vector_dim})
    mac_register_vector_dict = OrderedDict({'count': 0})

    performance_dict['subevent'] = OrderedDict({'multiplier_vector': multiplier_vector_dict,
                                                'accumulator_vector': accumulator_vector_dict,
                                                'register_vector': register_vector_dict,
                                                'mac_register_vector': mac_register_vector_dict})
    return performance_dict


# following two lines are used in testing
import sys, os, shutil

from loguru import logger

from archx.architecture import create_architecture_dict, save_architecture_dict, load_architecture_dict
from archx.utils import get_path, check_dict_equal


def test_create_architecture_dict():
    architecture_file = 'examples/mac_1_cycle/input/architecture/example.architecture.yaml'
    architecture_dict = create_architecture_dict(architecture_file)
    for module in architecture_dict:
        logger.info(module)
        print(module, architecture_dict[module], '\n')

    save_architecture_dict(architecture_dict, 'tests/test_architecture/example.architecture_dict.yaml')
    architecture_dict_loaded = load_architecture_dict('tests/test_architecture/example.architecture_dict.yaml')
    assert check_dict_equal(architecture_dict, architecture_dict_loaded)


def test_cleanup():
    path = get_path('tests/test_architecture/')
    shutil.rmtree(path)


if __name__ == "__main__":
    test_create_architecture_dict()
    test_cleanup()




from collections import OrderedDict
from archx.utils import get_prod

def router_event(router, architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    frequency = architecture_dict[router]['query']['frequency']

    router_width = architecture_dict[router]['instance'][-1]
    router_bandwidth = architecture_dict[router]['query']['bandwidth'] * 2**23

    cycle_count = 1 / router_width / router_bandwidth

    performance_dict['cycle_count'] = OrderedDict({'value': cycle_count, 'unit': 'cycle'})
    performance_dict['runtime'] = OrderedDict({'value': cycle_count / 1000 / frequency, 'unit': 'ms'})

    router_dict = OrderedDict({'count': 1})
    performance_dict['subevent'] = OrderedDict({router: router_dict})
    return performance_dict

def irouter_mapping(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    return router_event('irouter', architecture_dict, workload_dict)

def wrouter_mapping(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    return router_event('wrouter', architecture_dict, workload_dict)

def orouter_mapping(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    return router_event('orouter', architecture_dict, workload_dict)